#!/usr/bin/env python

"""This script produces a measure of the irregularity of
a waveform.  It is zero if the waveform is precisely periodic.
It is approximately 1 for white noise.

This is used as the aperiodicity measure in
Kochanski, Coleman, Grabe and Rosner JASA 2005.
"""

import sys
import math as M
import numpy
from gmisclib import die
from gpk_voicing import power
import gpkimgclass
from gpk_voicing import voice_misc

DELTA_DELAY = 1.0/(2400.0 * 2 * M.pi)
# DELTA_DELAY is chosen so that we can align the oscillations
# at the first formant (f1 is normally under 1200 Hz)
# to an accuracy of 1 radian or less.

Fmin = 50.0
Fmax = 500.0
Fhp = 500.0
WINDOW = 0.020

BLOCK_EDGE = 0.1	# Seconds

DBG = None

# def one_predict(d, delay, dt, Dt):
	# zl = Num.zeros((delay/2,), Num.Float)
	# zr = Num.zeros((delay-delay/2,), Num.Float)
	# dtmp = Num.concatenate( (zl, d, zr) )
	# tmp = dtmp[delay:] - dtmp[:-delay]
	# assert tmp.shape == d.shape
	# return power.smooth(tmp*tmp, dt, Dt, extra=WINDOW)


def predict_err(d, delays, dt, Dt):
	global WINDOW
	assert min(delays) >= 0
	mxdelay = max(delays)//2 + 1
	n = d.shape[0]
	ze = numpy.zeros((mxdelay,))
	dtmp = numpy.concatenate( (ze, d, ze) )
	del d	# save memory.
	o = None
	for delay in delays:
		dlh = delay//2
		dlH = delay - dlh
		ddtmp = dtmp[ mxdelay+dlh : mxdelay+dlh+n ] - dtmp[ mxdelay-dlH : mxdelay-dlH+n ]
		assert ddtmp.shape[0] == n, "Shapes do not match: %s and %s" % (ddtmp.shape, n)
		tmp, t0 = power.smooth(ddtmp**2, dt, Dt, extra=WINDOW)
		if DBG:
			DBG.plot(tmp)
		if o is None:
			o = tmp
		else:
			numpy.minimum(o, tmp, o)
	if DBG:
		DBG.show()
	return o


def local_power(d, dt, Dt):
	global WINDOW
	return power.smooth(d**2, dt, Dt, extra=WINDOW)


def fractional_step_range(initial, final, step=1):
	"""Generates an array of window offsets.
	No two offsets are equal, and they are integers
	between initial and final."""

	assert step != 0
	if step < 0:
		return fractional_step_range(initial, final, -step)
	if final < initial:
		return []
	o = []
	tmp = float(initial)
	assert tmp > 0.5, "Initial=%g too close to zero, or even negative!" % tmp
	ilast = 0
	while tmp <= final:
		itmp = int(round(tmp))
		if itmp != ilast and itmp>=initial:
			o.append(itmp)
		ilast = itmp
		tmp += step
	return o


def aperiodicity(d, dt, Dt):
	if Fmin*dt > 0.5 or Fmin*dt<1.0/d.shape[0]:
		die.warn("Silly time scale")
	die.info("Hipass at %f" % (Fmin*dt))
	d = voice_misc.hipass_sym_butterworth(d, Fmin*dt)
	d = voice_misc.hipass_first_order(d, Fhp*dt)
	if DBG is not None:
		DBG.plot(numpy.arange(d.shape[0])*dt, d, 'k-')
		DBG.title('filtered signal')
	p, tshift = local_power(d, dt, Dt)
	if DBG is not None:
		DBG.figure()
		DBG.plot(tshift+numpy.arange(p.shape[0])*Dt, p)
		DBG.title('Local power')
		DBG.show()
	assert tshift < Dt
	imin = 1.0/(dt * Fmax)
	imax = 1.0/(dt * Fmin)
	istep = DELTA_DELAY/dt
	delays = fractional_step_range(imin, imax, istep)
	perr = predict_err(d, delays, dt, Dt)
		# The factor of 2.0 below is because perr is a difference
		# of two signals.  If they were independent random noise,
		# then the variances would add, so that perr=2*p for
		# white noise.
		# In other words, this measure goes to 1 for white noise,
		# and to zero for a periodic signal.
	assert perr.shape == p.shape
	return (tshift, numpy.sqrt(perr/(2.0*p)))


def  block_aperiodicity(data, dt, Dt, blocksize=6.0):
	if blocksize is None:
		return aperiodicity(data, dt, Dt)
	assert BLOCK_EDGE >= Dt
	blocksize = Dt * int(round(blocksize/Dt))
	assert blocksize > BLOCK_EDGE
	m = data.shape[0]
	ntau = int(M.ceil(m*dt/blocksize))
	nout = int(M.ceil(m*dt/Dt))
	pad = int(round(Dt*round(BLOCK_EDGE/Dt)/dt))

	irr = None
	tshift = None
	S0 = 0.0
	e = 0
	for i in range(ntau):
		s0 = int(round((m*i)/float(ntau)))
		se = int(round((m*(i+1))/float(ntau)))
		ps0 = max(0, s0-pad)
		pse = min(m, se+pad)
		ttshift, x = aperiodicity(data[ps0:pse], dt, Dt)
		if tshift is None:
			tshift = ttshift
		if irr is None:
			irr = numpy.zeros((nout+1,))
		tau0 = int(M.floor((s0-ps0)*(dt/DT)))
		nS = min((se-s0)*(dt/DT), x.shape[0]-tau0)
		inS = int(M.ceil(nS))
		iS0 = int(M.floor(S0))
		assert iS0+inS <= nout, "S0=%.1f nS=%.1f S0+nS=%.1f nout=%d" % (S0, nS, S0+nS, nout)
		# print 'irr', irr.shape, iS0, iS0+inS, 'x=', x.shape, tau0, tau0+inS
		irr[iS0:iS0+inS] = x[tau0:tau0+inS]
		S0 += nS
		e = iS0+inS
	assert nout-2 <= e <= nout+1, "e=%d nout=%d" % (e, nout)
	return (tshift, irr[:e])


def test(pylab):
	import random
	DT = 0.01
	dt = 1.0/random.normalvariate(16000.0, 1000.0)
	L = random.expovariate(1.0/15.0)
	n = int(round(L/dt))
	signal = numpy.random.normal(0.0, 0.1, size=(n,))
	step = int(round(2.0/dt))
	k = int(round(0.5/dt))
	omega = 2*M.pi*100.0
	i = 0
	while i < n:
		kk = min(k, n-i)
		signal[i:i+kk] = numpy.sin(numpy.arange(kk)*dt*omega)
		if i+k < n:
			kk = min(k, n-i-k)
			signal[i+k:i+k+kk] = numpy.random.normal(0.0, 0.9, size=(kk,))
		i += step
	
	if pylab:
		pylab.plot(numpy.arange(signal.shape[0])*dt, signal)
	tshift, o = block_aperiodicity(signal, dt, DT)
	x = numpy.arange(o.shape[0])*DT + tshift
	if pylab:
		pylab.plot(x, o)
		pylab.show()



if __name__ == '__main__':
	DT = 0.01	# Seconds.  Default output sampling interval.
	arglist = sys.argv[1:]
	Blocksize = 10.0
	arglist0 = arglist
	column = None
	signalfile = None
	outfile = None
	extrahdr = {}
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-params':	# For optimization purposes
			Fmin = atof(arglist.pop(0))
			Fmax = atof(arglist.pop(0))
			Fhp = atof(arglist.pop(0))
			WINDOW = atof(arglist.pop(0))
		elif arg == '-dt':
			DT = float(arglist.pop(0))
		elif arg == '-f':
			signalfile = arglist.pop(0)
		elif arg == '-c':
			tmp = arglist.pop(0)
			try:
				column = int( tmp )
			except ValueError:
				column = tmp
		elif arg == '-b':
			Blocksize = float(arglist.pop(0))
			if Blocksize <= 0.0:
				Blocksize = None
		elif arg == '-d':
			import pylab as DBG
		elif arg == '-test':
			test(None)
			sys.exit(0)
		elif arg == '-Test':
			import pylab
			test(pylab)
			sys.exit(0)
		elif arg == '-o':
			outfile = arglist.pop(0)
		elif arg == '-write':
			extrahdr = dict( [q.strip().split('=', 1)
					for q in arglist.pop(0).split(';') ]
					)
		else:
			die.info("Unrecognized flag: %s" % arg)
			print __doc__
			die.exit(1)
	if arglist and signalfile is None:
		signalfile = arglist.pop(0)
	if arglist and outfile is None:
		outfile = arglist.pop(0)
	elif outfile is None:
		outfile = 'pdur.dat'
	if column is None:
		column = 0
	if signalfile is None:
		die.info("No signal file specified.")
		print __doc__
		die.exit(1)
	if arglist:
		die.info('Extra arguments!')
		print __doc__
		die.exit(1)
	signal = gpkimgclass.read(signalfile)
	try:
		signal.column(column)
	except KeyError:
		die.info("File has %d columns" % signal.n[1])
		die.die("Bad column: %s" % str(column))

	tshift, o = block_aperiodicity(signal.column(column), signal.dt(), DT,
					blocksize=Blocksize)

	hdr = signal.hdr.copy()
	hdr['program'] = sys.argv[0]
	hdr['ARGV'] = arglist0
	hdr['input_file'] = signalfile
	hdr['column'] = column
	hdr['CDELT2'] = DT
	hdr['CRPIX2'] = 1
	hdr['CRVAL2'] = signal.start() + tshift
	hdr['CRPIX1'] = 1
	hdr['CDELT1'] = 1
	hdr['TTYPE1'] = 'aperiodicity'
	hdr['BITPIX'] = -32
	hdr['DataSamplingFreq'] = 1.0/signal.dt()
	hdr.update( extrahdr )
	gpkimgclass.gpk_img(hdr, o).write(outfile)




import numpy
from gmisclib import die
from gmisclib import wavio
from gmisclib import Numeric_gpk as NG

NEARCLIP = 32500

def wav_read(filename, channel=None, info={}):
	"""Quick and dirty way to read a single channel of audio from a
	multichannel file.  If the channel is not specified, either as an
	argument or as info['Channel'], then it will pick whatever channel
	has the largest amplitude.   Needless to say, this is not entirely
	reliable unless you know that the other channel is silent.
	@param channel: which channel to use?  It can be a string name, an integer column number
		or None.
	@return: a gpk_img data structure holding the audio and header information.
	@param info: a dictionary with miscellaneous metadata.   Only C{info['Channel']} is used.
	@type info: C{dict}.
	@type channel: C{string}, C{int}, or C{None}.
	"""
	if channel is None:
		channel = info.get('Channel', None)

	x = wavio.read(filename)
	trouble = []
	if channel is None and x.d.shape[1] > 1:
		assert x.d.shape[1] < 1000
		rms = NG.block_stdev(x.d)
		biggest = numpy.argmax(rms)
		x.d = x.d.take([biggest], axis=1)
		assert len(x.d.shape) == 2
		die.info('Taking data from channel %d' % biggest)
		if rms[1-biggest] > 0.1*rms[biggest]:
			die.warn('Warning: channel %d is comparable: MAD=%.2g/%.2g'
					% (1-biggest, rms[1-biggest], rms[biggest])
					)
			x.hdr['WhyAudioChannel'] = 'It was slightly louder'
			trouble.append('Ambiguous channel')
		else:
			x.hdr['WhyAudioChannel'] = 'It was much louder'
	elif channel is None and x.d.shape[1]==1:
		x.hdr['WhyAudioChannel'] = 'There was only one'
	else:
		x.d = x.d.take([channel], axis=1)
		x.hdr['WhyAudioChannel'] = 'Specified in g_audioprep.py'

	if NG.N_maximum(x.column(0)) > NEARCLIP or NG.N_minimum(x.column(0))<-NEARCLIP:
		da = numpy.absolute(x.column(0))
		die.warn("Data is near clipping: maximum abs(value)=%d, %.1f%% near clipping"
			% (NG.N_maximum(da), numpy.sum(numpy.greater(da, NEARCLIP))/float(da.shape[0])))
		trouble.append('Near clipping')

	x.hdr['Trouble'] = ','.join(trouble)
	x.hdr['AudioChannel'] = channel
	assert x.d.shape[1] == 1
	return x



#!/usr/bin/env python

"""Duration estimator for speech.

Usage: s_slope [flags]
Flags: -f FFF opens file FFF for reading.
	-c XXX  selects column XXX (either a data name or an integer).
	-o FFF  sets the output file.
	-dt #.##   Sets the interval for calculating the output.

It takes a local spectrum,
bins it onto the Bark scale,
converts to perceptual loudness (via **E).
Then, it computes a measure of how far you can go
from each point before the spectrum changes too much.
"""

import sys
import math as M
import numpy
from gmisclib import die
import gpk_voicing.fv_misc as FVM
import gpkimgclass
from gpk_voicing import pseudo_dur as PD
from gpk_voicing import fv_pdur as FVP

C = 10.2
Lfac = 1.03
TYP_DUR = 0.1	# Seconds

def pdur(data, dt, Dt, out, c=None, lfac=None):
	# print 's.n=', data.shape
	sp, descr, t0 = FVP.feature_vec(data, dt, Dt, Nsv=FVM.NSV)
	nfvc = len(sp)
	sp = numpy.transpose(sp)
	assert sp.shape[1] == nfvc

	ns = sp.shape[0]
	assert abs(ns*Dt - data.shape[0]*dt) < 0.1*ns*Dt
	o = numpy.zeros((ns,), numpy.float)
	for i in range(ns):
		plp, pcp = PD.pdur_guts(sp, i, 1, Dt, c)
		plm, pcm = PD.pdur_guts(sp, i, -1, Dt, c)
		if out=='pseudoduration':
			o[i] = lfac * (plp + plm)
		elif out=='log(pseudoduration)':
			o[i] = M.log((lfac/TYP_DUR) * (plp + plm))
		elif out=='center_time':
			o[i] = Dt*(pcp + pcm)/(plp+plm)
		else:
			die.die('Whoops: out=%s' % out)
	return o


if __name__ == '__main__':
	# try:
		# import psyco
		# psyco.full()
	# except ImportError:
		# pass

	DT = 0.01	# Seconds.  Default output sampling interval.
	arglist = sys.argv[1:]
	arglist0 = arglist
	column = None
	signalfile = None
	outfile = None
	extrahdr = {}
	RTYPE = 'pseudoduration'
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-DT' or arg == '-dt':
			DT = float(arglist.pop(0))
		elif arg == '-f':
			signalfile = arglist.pop(0)
		elif arg == '-c' or arg == '-col':
			tmp = arglist.pop(0)
			try:
				column = int( tmp )
			except ValueError:
				column = tmp
			die.note("signalfile", signalfile)
		elif arg == '-lfac':
			LFAC = M.exp(float(arglist.pop(0)))
		elif arg == '-C':
			C = float(arglist.pop(0))
		elif arg == '-dur':
			RTYPE = 'pseudoduration'
		elif arg == '-logdur':
			RTYPE = 'log(pseudoduration)'
		elif arg == '-ctr':
			RTYPE = 'center_time'
		elif arg == '-o':
			outfile = arglist.pop(0)
		elif arg == '-write':
			extrahdr = dict( [q.strip().split('=', 1)
					for q in arglist.pop(0).split(';') ]
					)
		else:
			die.info("Unrecognized flag: %s" % arg)
			print __doc__
			die.exit(1)
	if arglist and signalfile is None:
		signalfile = arglist.pop(0)
	if arglist and outfile is None:
		outfile = arglist.pop(0)
	elif outfile is None:
		outfile = 'pdur.dat'
	if column is None:
		column = 0
	if signalfile is None:
		die.info("No signal file specified.")
		print __doc__
		die.exit(1)
	if arglist:
		die.info('Extra arguments!')
		print __doc__
		die.exit(1)
	signal = gpkimgclass.read(signalfile)
	try:
		signal.column(column)
	except KeyError:
		die.info("File has %d columns" % signal.n[1])
		die.die("Bad column: %s" % str(column))

	try:
		o = pdur(signal.column(column), signal.dt(), DT, RTYPE, c=C, lfac=Lfac)
	except:
		die.catch("Exception in pdur calculation")
		raise

	hdr = signal.hdr.copy()
	hdr['program'] = sys.argv[0]
	hdr['ARGV'] = arglist0
	hdr['input_file'] = signalfile
	hdr['column'] = column
	hdr['CDELT2'] = DT
	hdr['CRPIX2'] = 1
	hdr['CRVAL2'] = signal.start()
	hdr['CDELT1'] = 1
	hdr['TTYPE1'] = RTYPE
	hdr['BITPIX'] = -32
	hdr['DataSamplingFreq'] = 1.0/signal.dt()
	hdr.update( extrahdr )
	gpkimgclass.gpk_img(hdr, o).write(outfile)



#!/usr/bin/env python

import os
import sys
import math
import gpkimgclass
from gmisclib import Num
from gmisclib import wavio
from gmisclib import gpkmisc
from gmisclib import die
from gpk_voicing import emphasis

DT = 0.03
R = 1
D = 1

if __name__ == '__main__':
	import sys
	arglist = sys.argv[1:]
	outfile = 'foo.wav'
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-stereo':
			D = 2
		elif arg == '-mono':
			D = 1
		elif arg == '-resample':
			R = int(arglist.pop(0))
		elif arg == '-o':
			outfile = arglist.pop(0)
		else:
			die.die('Unrecognized flag: %s' % arg)

	normfacs = []
	data = []
	for f in arglist:
		gotit = False
		try:
			q = gpkimgclass.read(f)
			gotit = True
		except gpkimgclass.CannotReadDataFile, x:
			die.info('File %s unreadable: %s' % (f, str(x)))
		if not gotit:
			q = wavio.read(f)

		n = 0
		s = 0.0
		for c in range(q.d.shape[1]):
			tmp = emphasis.simple_emphasis(q.column(c), q.dt(), DT)
			tt = Num.sort(Num.ravel(tmp))
			assert len(tt.shape) == 1
			if tt.shape[0] > 0:
				s += tt[int(tt.shape[0]*0.9)]
				n += 1
		assert n > 0
		# Normalize amplitudes:
		print "Loudness=", s/n
		normfac = math.pow(n/s, 1.5)
		q.d = q.d * normfac
		data.append(q)
		normfacs.append(normfac)

	normfac = gpkmisc.median(normfacs)
	for q in data:
		q.d /= normfac

	o = []
	for q in data:
		assert abs(math.log(q.dt()/data[0].dt())) < 0.0001
		if q.d.shape[1] == D:
			o.append(q.d[::R])	# Resample
		elif q.d.shape[1] == 1:		# Mono->Stereo
			tmp = Num.transpose(Num.array([q.d[::R,0]]*D))
			print "ts=", tmp.shape, "qds=", q.d.shape
			assert tmp.shape[0] == (q.d.shape[0]+R-1)/R
			assert tmp.shape[1] == D
			o.append(tmp)
		elif D == 1:		# Stereo -> Mono
			die.die('Unimplemented')
	da = Num.concatenate(o, axis=0)

	header = {'CDELT2': R*data[0].dt(), 'CRPIX2':0, 'CRVAL2':0.0}
	if outfile.endswith('.wav'):
		unwritten = True
		while unwritten:
			try:
				wavio.write(gpkimgclass.gpk_img(header, da), outfile)
				unwritten = False
			except ValueError, x:
				assert 'overflow' in str(x)
				Num.divide(da, 1.5, da)
				die.warn("Overflow on writing to .wav file.  Scaling down by factor of 1.5.")
	else:
		gpkimgclass.gpk_img(header, da).write(outfile)



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#============================================================
#GAMMATONE_C: A C implementation of the 4th order gammatone filter
#-------------------------------
#  [bm, env, instp, ] = gammatone_c(x, fs, cf, hrect) 
#
#  x     - input signal
#  fs    - sampling frequency (Hz)
#  cf    - centre frequency of the filter (Hz)
#  hrect - half-wave rectifying if hrect = 1 (default 0)
#
#
#  bm    - basilar membrane displacement
#  env   - instantaneous envelope
#  instp - instantaneous phase (unwrapped radian)
#
#  >> mex gammatone_c.c
#  >> bm = gammatone_c(x, 16000, 200);
#
# This was translated from MATLAB to python by Greg Kochanski.
# More detail on the original can be found at:
#  http://www.dcs.shef.ac.uk/~ning/resources/gammatone/
# Original author:
#  Ning Ma, University of Sheffield
#  n.ma@dcs.shef.ac.uk, 09 Mar 2006
# 

import math
import cython
import numpy
cimport numpy
ctypedef numpy.float_t DTYPE

cdef extern from "math.h":
	double floor(double)
	double exp(double)
	double sqrt(double)
	double sin(double x)
	double cos(double x)
	double fabs(double x)
	double atan2(double y, double x)

# /*=======================
 # * Useful Const
 # *=======================
cdef double BW_CORRECTION = 1.0190
cdef double M_PI = math.pi

# /*=======================
 # * Utility functions
 # *=======================

cdef double myMod(double x, double y):
	return x - y*floor(x/y)

cdef double erb(double x):
	return 24.7*(4.37e-3*x+1.0)

M_BASILAR = 0
M_BASILAR_RECT = 1
M_AMP_PHASE = 2

#  [bm, env, instp, ] = gammatone_c(x, fs, cf, hrect) 
@cython.boundscheck(False)
def gammatone(numpy.ndarray[DTYPE, ndim=1, mode="c"] x, double dt, double cf, int mode):
	assert mode in [M_BASILAR, M_BASILAR_RECT, M_AMP_PHASE]
	cdef int hmode = mode
	cdef int nsamples = x.shape[0]
#/*=========================================
# * output arguments
# *=========================================
	cdef numpy.ndarray[DTYPE, ndim=1, mode="c"] bm
	cdef numpy.ndarray[DTYPE, ndim=2, mode="c"] env_phase
	if hmode==0 or hmode==1:
		bm = numpy.zeros((nsamples,))
	else:
		env_phase = numpy.zeros((nsamples,2))

#/*=========================================
# * Initialising variables
# *=========================================
	cdef double oldphase = 0.0
	cdef double tpt = ( M_PI + M_PI ) * dt
	cdef double tptbw = tpt * erb ( cf ) * BW_CORRECTION
	cdef double a = exp ( -tptbw )

#/* based on integral of impulse response */
	cdef double gain = ( tptbw*tptbw*tptbw*tptbw ) / 3.0

#/* Update filter coefficiants */
	cdef double a1 = 4.0*a
	cdef double a2 = -6.0*a*a
	cdef double a3 = 4.0*a*a*a
	cdef double a4 = -a*a*a*a
	cdef double a5 = 4.0*a*a
	cdef double p0r = 0.0
	cdef double p1r = 0.0
	cdef double p2r = 0.0
	cdef double p3r = 0.0
	cdef double p4r = 0.0
	cdef double p0i = 0.0
	cdef double p1i = 0.0
	cdef double p2i = 0.0
	cdef double p3i = 0.0
	cdef double p4i = 0.0
 
#/*===========================================================
# * exp(a+i*b) = exp(a)*(cos(b)+i*sin(b))
# * q = exp(-i*tpt*cf*t) = cos(tpt*cf*t) + i*(-sin(tpt*cf*t))
# * qcos = cos(tpt*cf*t)
# * qsin = -sin(tpt*cf*t)
# *===========================================================
	cdef double coscf = cos ( tpt * cf )
	cdef double sincf = sin ( tpt * cf )
	cdef double qcos = 1.0	#   /* t=0 & q = exp(-i*tpt*t*cf)*/
	cdef double qsin = 0.0
	cdef double oldcs, oldsn, dp, dps, u0r, u0i
	cdef int t
	for t in range(nsamples):
		#/* Filter part 1 & shift down to d.c. */
		p0r = qcos*x[t] + a1*p1r + a2*p2r + a3*p3r + a4*p4r
		p0i = qsin*x[t] + a1*p1i + a2*p2i + a3*p3i + a4*p4i

		#/* Filter part 2 */
		u0r = p0r + a1*p1r + a5*p2r
		u0i = p0i + a1*p1i + a5*p2i

		#/* Update filter results */
		p4r = p3r
		p3r = p2r
		p2r = p1r
		p1r = p0r
		p4i = p3i
		p3i = p2i
		p2i = p1i
		p1i = p0i
  
		#/*==========================================
		#* Basilar membrane response
		#* 1/ shift up in frequency first: (u0r+i*u0i) * exp(i*tpt*cf*t)
		#				= (u0r+i*u0i) * (qcos + i*(-qsin))
		#* 2/ take the real part only: bm = real(exp(j*wcf*kT).*u) * gain;
		#*==========================================
		if hmode==0 or hmode==1:
			bm[t] = ( u0r * qcos + u0i * qsin ) * gain
			if hmode==1 and bm[t] < 0:
				bm[t] = 0.0	#  /* half-wave rectifying */

		#/*==========================================
		#* Instantaneous Hilbert envelope
		#* env = abs(u) * gain;
		#*==========================================
		if hmode==2:
			env_phase[t,0] = sqrt ( u0r * u0r + u0i * u0i ) * gain

		#/*==========================================
		#* Instantaneous phase
		#* instp = unwrap(angle(u));
		#*==========================================
			env_phase[t,1] = atan2 ( u0i, u0r )
			# /* unwrap it */
			dp = env_phase[t,1] - oldphase
			if fabs ( dp ) > M_PI:
				dps = myMod ( dp + M_PI, 2 * M_PI) - M_PI;
				if dps == -M_PI and dp > 0:
					dps = M_PI
				env_phase[t,1] = env_phase[t,1] + dps - dp
			oldphase = env_phase[t,1];

		# /*====================================================
		# * The basic idea of saving computational load:
		# * cos(a+b) = cos(a)*cos(b) - sin(a)*sin(b)
		# * sin(a+b) = sin(a)*cos(b) + cos(a)*sin(b)
		# * qcos = cos(tpt*cf*t) = cos(tpt*cf + tpt*cf*(t-1))
		# * qsin = -sin(tpt*cf*t) = -sin(tpt*cf + tpt*cf*(t-1))
		# *====================================================
		oldcs = qcos
		oldsn = qsin
		qcos = coscf * oldcs + sincf * oldsn
		qsin = coscf * oldsn - sincf * oldcs

	if hmode==0 or hmode==1:
		return bm
	return env_phase





@cython.boundscheck(False)
def hipass_first_order(numpy.ndarray[DTYPE, ndim=1, mode="c"] d, double cutoff_freq):
	"""A first-order high-pass filter.
	Cutoff freq is measured in cycles per point."""

	assert cutoff_freq >= 0.0
	cdef int n = d.shape[0]
	cdef numpy.ndarray[DTYPE, ndim=1, mode="c"] o = numpy.zeros((n,))
	cdef double tau = 1.0/(2*M_PI*cutoff_freq)
	cdef double alpha = tau/(tau+1.0)
	cdef int i
	for i in range(1,n):
		o[i] = alpha * (o[i-1] + (d[i] - d[i-1]))
	return o


@cython.boundscheck(False)
def lopass_first_order(numpy.ndarray[DTYPE, ndim=1, mode="c"] d, cutoff_freq):
	"""A first-order low-pass filter.
	Cutoff freq is measured in cycles per point."""
	assert cutoff_freq >= 0.0
	cdef int n = d.shape[0]
	cdef numpy.ndarray[DTYPE, ndim=1, mode="c"] o = numpy.zeros((n,))
	cdef double tau = 1.0/(2*M_PI*cutoff_freq)
	cdef double alpha = 1.0/(tau+1.0)
	cdef int i
	for i in range(1,n):
		o[i] = o[i-1] + alpha * (d[i]-o[i-1])
	return o



import math
import cython
import numpy
cimport numpy
from cython import wraparound

from gmisclib import hilbert_xform

ctypedef numpy.float_t DTYPE

cdef extern from "math.h":
	double floor(double)
	double ceil(double)
	double cos(double)

cdef extern from "gpklib.h":
	int roundI(double)



cdef int max(int a, int b):
	if a > b:
		return a
	return b

cdef int min(int a, int b):
	if a < b:
		return a
	return b


@cython.boundscheck(False)
def smooth_guts(numpy.ndarray[DTYPE, ndim=1, mode="c"] ph,
		double dt_in, double dt_out, double w, wt=None):
	assert w >= 1.0
	assert dt_in <= dt_out
	cdef double dt_ratio = dt_out/dt_in
	cdef int no = <int>(floor(ph.shape[0]/dt_ratio))
	cdef numpy.ndarray[DTYPE, ndim=1, mode="c"] o = numpy.zeros((no,))
	cdef numpy.ndarray[DTYPE, ndim=1, mode="c"] owt, iwt, window
	if wt is not None:
		owt = numpy.zeros((no,), numpy.float)
		iwt = numpy.as_array(wt)

	cache = {}

	cdef double SS = float(10.0)
	cdef int ni = ph.shape[0]
	cdef int i, j0, je, j
	cdef double jr, jrr, s, sw, wtmp
	for i in range(no):
		jr = i*dt_ratio
		jrr = roundI(jr*SS)/SS
		j0 = max(0, <int>(ceil(jrr-w)))
		je = min(ni, <int>(floor(jrr+w))+1)
		assert je > j0
		wkey = (je-j0, jrr-j0)
		try:
			window = cache[wkey]
		except KeyError:
			window = _wfunc(je-j0, jrr-j0, w)
			cache[wkey] = window

		s = 0.0
		sw = 0.0
		if wt is not None:
			for j in range(je-j0):
				wtmp = iwt[j0+j]*window[j]
				sw += wtmp
				s += wtmp * ph[j0+j]
		else:
			for j in range(je-j0):
				s += ph[j0+j]*window[j]
				sw += window[j]
		o[i] = s/sw
	if wt is None:
		return o, None
	return o, owt



@cython.boundscheck(False)
def _wfunc(int n, double ctr, double w):
	cdef int i
	cdef double phase
	cdef double piw = math.pi/w
	cdef numpy.ndarray[DTYPE, ndim=1, mode="c"] o = numpy.zeros((n,))
	for i in range(n):
		phase = piw*(i-ctr)
		assert phase>=-3.142 and phase<=3.142, "Bad phases: %s or %s for %s %s %s" % (str(phase), str(phase), str(n), str(ctr), str(w))
		o[i] = 1.0 + cos(phase)
	return o




import cython
import numpy
cimport numpy
ctypedef numpy.float_t DTYPE

cdef extern from "math.h":
	double fabs(double x)



def list_accum_abs_sub_dly(tk, dly):
	cdef int lng = tk[0].shape[0]
	cdef int dlyn = dly//2
	cdef int dlyp = dly - dlyn
	cdef numpy.ndarray[DTYPE, ndim=1, mode="c"]o = numpy.zeros((lng,))
	cdef numpy.ndarray[DTYPE, ndim=1, mode="c"] tkd
	cdef double a, b
	cdef int i
	for tkd in tk:
		for i in range(dlyn,lng-dlyp):
			a = tkd[i+dlyp]
			b = tkd[i-dlyn]
			o[i] += fabs(a-b)
	return o




#!/usr/bin/env python

"""Duration estimator for speech.

Usage: s_slope [flags]
Flags: -f FFF opens file FFF for reading.
	-c XXX  selects column XXX (either a data name or an integer).
	-o FFF  sets the output file.
	-dt #.##   Sets the interval for calculating the output.

It takes a local spectrum,
bins it onto the Bark scale,
converts to perceptual loudness (via **E).
Then, it computes a measure of how far you can go
from each point before the spectrum changes too much.
"""

import math as M
import numpy


def pdur_guts(s, t, dir, Dt, C):
	"""t is an integer; an index into the data.
	S is the normalized perceptual spectrum."""
	n = len(s)
	assert s.shape[1] < 200, "Implausibly long feature vector.  Is s transposed?"
	sumdiff = 0.0
	len_sum = 0.5	# The integral for the i==t case is trivial.
	ctr_sum = 0.125*dir	# The integral for i==t is trivial.
	i = t + dir
	while i>=0 and i<n and sumdiff<8:
		delta_diff = Dt * C * numpy.sum( numpy.absolute(s[i]-s[t])**2 )

		# We assume that the spectrum is s[i] in the
		# region from i-0.5 to i+0.5, and compute the
		# integral( M.exp(-integral((s(t0)-s(t''))**2 from 0 to t', dt''))
		# 	from 0 to i+0.5, dt')
		# The constancy of s over the interval (i-0.5,i+0.5)
		# makes the inner integral piecewise linear.
		# Sumdiff holds integral((s(t0)-s(t''))**2 from 0 to i-0.5)
		if delta_diff <= 0:
			# assume delta_diff = 0
			lsd = M.exp(-sumdiff)
			len_sum += lsd
			# integral( (t''-t0)*M.exp(-integral((s(t0)-s(t''))**2 from 0 to t', dt''))
			# 	from 0 to i+0.5, dt')
			slopeint = M.exp(-sumdiff) * 0.5
			ctr_sum += (i-t-0.5*dir)*lsd + dir * slopeint
			i += dir
		else:
			lsd = M.exp(-sumdiff) * (1.0-M.exp(-delta_diff))/delta_diff
			len_sum += lsd
			# integral( (t''-t0)*M.exp(-integral((s(t0)-s(t''))**2 from 0 to t', dt''))
			# 	from 0 to i+0.5, dt')
			slopeint = M.exp(-sumdiff)	\
					* (1.0-M.exp(-delta_diff)*(1+delta_diff))/delta_diff**2
			ctr_sum += (i-t-0.5*dir)*lsd + dir * slopeint
			i += dir
			sumdiff += delta_diff
		# Now, sumdiff holds integral((s(t0)-s(t''))**2 from 0 to i+0.5)
	# print 't=', t, sum
	return (len_sum*Dt, ctr_sum*Dt)





#!/usr/bin/env python

import math
import numpy
from gmisclib import hilbert_xform
from gpk_voicing import power1 as P1

# import pylab

def calc_var_of_window():
	N = 100
	phase = (math.pi/float(N))*(numpy.arange(N)+0.5)
	w = 1 + numpy.cos(phase)
	return numpy.sum(w*phase**2)/numpy.sum(w)

_ALPHA = 1.0/math.sqrt(calc_var_of_window())


def smooth(ph, dt_in, dt_out, extra=0.0, wt=None):
	"""Smooths a data set, simultaneously resampling to
	a lower sampling rate.   It uses successive boxcar averages
	followed by decimations for the initial smooth, then a convolution
	with a Gaussian.   Even if C{dt_out>>dt_in}, it only uses
	C{O[log(dt_out/dt_in)} operations.
	@param dt_in: input sampling rate.
	@param dt_out: output sampling rate.
	@param extra: extra smoothing time constant to apply.
		Extra is the standard deviation of a Gaussian kernel smooth that is
		applied as the last step.
		This last step is not implemented efficiently, so if if C{extra>>dt_out}
		it can slow down the algorithm substantially.
	@type extra: float  (in the same units as dt_in and dt_out).
	@type dt_in: float  (in the same units as extra and dt_out).
	@type dt_out: float  (in the same units as dt_in and extra).
	@param ph: Normally a 1-dimensional array containing data to be smoothed.
		If the data is higher-dimensional, the time axis is assumed to run
		along axis=0, and the return value will be an array of the same dimension.
	@type ph: L{numpy.ndarray}.
	@param ph: None (which indicates a uniform weighting) or
		a L{numpy.ndarray} that is the same length (axis 0) as C{ph}.
	@type wt: L{numpy.ndarray}
	@return: C{(rv, t0)} where C{rv} is a L{numpy} array and C{t0} it a C{float}
		offset of the first element, relative to the start of the input data.
	"""
	assert dt_in <= dt_out
	dt_in = float(dt_in)
	dt_ratio = float(dt_out)/dt_in
	w = math.hypot(dt_out, _ALPHA*extra)
	# print "w=", w, "dt_ratio=", dt_ratio, "ALPHA=", _ALPHA, "extra=", extra, "dt_in=", dt_in
	assert w >= dt_in

	# First, we try to see if it makes sense to do it in three
	# passes:
	dt_m1 = min(dt_out, ((w/dt_in)**0.667)*dt_in)
	ifac = int(round(dt_out/dt_m1))
	dt_m1 = dt_out/ifac
	assert dt_in <= dt_m1 <= dt_out
	dt_m2 = dt_m1/int(round(math.sqrt(dt_m1/dt_in)))
	assert dt_in <= dt_m2 <= dt_m1
	if w >= 3*dt_m1 and dt_m1 >= 3*dt_m2 and dt_m2 >= 3*dt_in:
		# print '3 steps:', dt_in, dt_m2, dt_m1, dt_out
		# pylab.plot(numpy.arange(ph.shape[0])*dt_in, ph)
		# print "dt_m1/dt_m2=", dt_m2/dt_in
		sm2, wt2 = P1.smooth_guts(ph, dt_in, dt_m2, dt_m2/dt_in, wt)
		# pylab.plot(numpy.arange(sm2.shape[0])*dt_m2, sm2)
		# print "dt_m1/dt_m2=", dt_m1/dt_m2
		sm1, wt1 = P1.smooth_guts(sm2, dt_m2, dt_m1, dt_m1/dt_m2, wt2)
		# pylab.plot(numpy.arange(sm1.shape[0])*dt_m1, sm1)
		# print "dt_m1=", dt_m1, "dt_out=", dt_out, "w/dt_m1=", w/dt_m1
		tmp = P1.smooth_guts(sm1, dt_m1, dt_out, w/dt_m1, wt1)[0]
		# pylab.plot(numpy.arange(tmp.shape[0])*dt_out, tmp)
		return (tmp, 0.0)
		# return (P1.smooth_guts(sm1, dt_m1, dt_out, w/dt_m1, wt1)[0], 0.0)
	# If not, we try to see if it makes sense to do it in two
	# passes:
	dt_mid = min(dt_out, math.sqrt(w*dt_in))
	ifac = int(round(dt_out/dt_mid))
	dt_mid = dt_out/ifac
	assert dt_in <= dt_mid <= dt_out
	if w >= 3*dt_mid and dt_mid >= 3*dt_in:
		# print '2 steps'
		sm1, wt1 = P1.smooth_guts(ph, dt_in, dt_mid, dt_mid/dt_in, wt)
		return (P1.smooth_guts(sm1, dt_mid, dt_out, w/dt_mid, wt1)[0], 0.0)
	# Otherwise, we do it in a single pass.
	# print '1 step'
	return (P1.smooth_guts(ph, dt_in, dt_out, w/dt_in, wt)[0], 0.0)


def smooth_guts(ph, dt_in, dt_out, w, wt=None):
	assert w >= 1.0
	assert dt_in <= dt_out
	dt_ratio = dt_out/dt_in
	no = int(math.floor(ph.shape[0]/dt_ratio))
	o = numpy.zeros((no,)+ph.shape[1:], numpy.float)
	if wt is not None:
		owt = numpy.zeros((no,)+ph.shape[1:], numpy.float)

	# print 'w=', w, 'dj=', dj, 'dt_ratio=', dt_ratio
	tmp0 = numpy.zeros((int(2*w)+3,))
	cache = {}
	SS = float(10.0)
	ni = ph.shape[0]
	for i in range(no):
		jr = i*dt_ratio
		jrr = round(jr*SS)/SS
		j0 = max(0, int(math.ceil(jrr-w)))
		je = min(ni, int(math.floor(jrr+w))+1)
		wkey = (je-j0, jrr-j0)
		try:
			window, wsum = cache[wkey]
		except KeyError:
			window = _wfunc(je-j0, jrr-j0, w)
			wsum = numpy.sum(window)
			cache[wkey] = (window, wsum)
		tmp = tmp0[:je-j0]
		if wt is not None:
			numpy.multiply(wt[j0:je], window, tmp)
			# owt[i] = numpy.sum(wt[j0:je]*window, axis=0)
			owt[i] = numpy.sum(tmp)
			numpy.multiply(tmp, ph[j0:je], tmp)
			# o[i] = numpy.sum(ph[j0:je]*wt[j0:je]*window, axis=0)/owt[i]
			o[i] = numpy.sum(tmp)/owt[i]
		else:
			numpy.multiply(ph[j0:je], window, tmp)
			# o[i] = numpy.sum(ph[j0:je]*window, axis=0)/numpy.sum(window, axis=0)
			o[i] = numpy.sum(tmp)/wsum
			owt = None
	return o, owt


def _wfunc(n, ctr, w):
	phase = (math.pi/w) * (numpy.arange(n) - ctr)
	# print 'phases for', n, ctr, w, '=', phase
	assert phase[0]>=-3.142 and phase[-1]<=3.142, "Bad phases: %s or %s for %s %s %s" % (str(phase[0]), str(phase[-1]), str(n), str(ctr), str(w))
	return 1.0 + numpy.cos(phase)


def old_smooth(ph, dt_in, dt_out, extra=0.0, wt=None):
	"""Smooths a data set, simultaneously resampling to
	a lower sampling rate.   It uses successive boxcar averages
	followed by decimations for the initial smooth, then a convolution
	with a Gaussian.   Even if C{dt_out>>dt_in}, it only uses
	C{O[log(dt_out/dt_in)} operations.
	@param dt_in: input sampling rate.
	@param dt_out: output sampling rate.
	@param extra: extra smoothing time constant to apply.
		Extra is the standard deviation of a Gaussian kernel smooth that is
		applied as the last step.
		This last step is not implemented efficiently, so if if C{extra>>dt_out}
		it can slow down the algorithm substantially.
	@type extra: float  (in the same units as dt_in and dt_out).
	@type dt_in: float  (in the same units as extra and dt_out).
	@type dt_out: float  (in the same units as dt_in and extra).
	@param ph: a 1???-dimensionan array containing data to be smoothed.
		(Query: will this work for higher-dimensional data?)
	@type ph: L{numpy.array}.
	@return: C{(rv, t0)} where C{rv} is a L{numpy} array and C{t0} it a C{float}
		offset of the first element, relative to the start of the input data.
	"""
	assert dt_in <= dt_out
	dt_ratio = dt_out/dt_in
	no = int(math.floor(ph.shape[0]/dt_ratio))
	# print 'no=', dt_in, len(ph), dt_out, dt_in*len(ph)/dt_out
	dti = dt_in
	# In vari, we accumulate the width of the equivalent
	# convolution kernel that has been applied so far.
	# Specifically, it is the second moment of the kernel.
	if wt is not None:
		ph = ph * wt
		weight = wt
	else:
		weight = numpy.ones((ph.shape[0],), numpy.int32)
	kwi = 0.0
	t0 = 0.0
	while dti < 0.33*dt_out:
		# print "ph=", ph
		nph = numpy.array(ph[::2], copy=True)
		ph12 = ph[1::2]
		numpy.add(nph[:ph12.shape[0]], ph12, nph[:ph12.shape[0]])
		del ph12
		ph = nph
		nwt = numpy.array(weight[::2], copy=True)
		wt12 = weight[1::2]
		numpy.add(nwt[:wt12.shape[0]], wt12, nwt[:wt12.shape[0]])
		del wt12
		weight = nwt

		t0 += 0.5*dti
		kwi += (dti/2.0)**2
		dti *= 2.0
	# kwo is the desired width of the convolution kernel
	# between th input and output data streams.
	# It is just the second moment of a rectangular boxcar average.
	kwo = extra**2 + dt_out**2/12.0 - dt_in**2/12.0
	# kwx is the extra smoothing kernel that is yet to be applied.
	kwx = kwo - kwi
	assert kwx >= 0.0

	# print 'ph=', ph
	# Now, we set up a Gaussian of the appropriate width
	# to do the final convolution.
	if kwx > 0.0:
		N = int(math.ceil(2.3*math.sqrt(kwx)/dti))
		i = numpy.arange(2*N+1) - N
		k0 = numpy.exp( (-1/(2*kwx)) * (i*dti)**2 )
		assert k0.shape[0] < ph.shape[0]
		kernel = k0/numpy.sum(k0)
		ph = numpy.convolve(ph, kernel, 1)
		if wt is not None:
			weight = numpy.convolve(weight, kernel, 1)
	numpy.divide(ph, weight, ph)
	# print 'phs=', ph

	# Then grab the necessary samples.
	q = numpy.arange(no)*(dt_out/dti)
	qi = numpy.around(q).astype(numpy.int)
	assert qi[-1] < ph.shape[0]
	rv = numpy.take(ph, qi)
	assert rv.shape[0] == no
	return (rv, t0)


def local_power(d, dt_in, dt_out, extra=0.0):
	""" THIS IS WRONG!   IT PROBABLY SHOULD BE something like
	sqrt(d**2 + hilbert_transform(d)**2).
	The hilbert transform just supplies the imaginary part of
	an analytic function.   Current code leaves out the
	read part!
	"""
	raise RuntimeError, "Bug!"
	cutoff = math.hypot(dt_out, extra)/dt_in
	ph = numpy.absolute(hilbert_xform.hilbert(d, cutoff))
	return smooth(ph, dt_in, dt_out, extra)


def test1():
	x = numpy.zeros((50,), numpy.float)
	x[10:] = 1.0
	k = [0.1, 0.2, 0.4, 0.2, 0.1]
	xk = numpy.convolve(x, k, 1)
	xs0, ts0 = smooth(x, 1.0, 1.0, 0.01)
	xs1, ts1 = smooth(x, 1.0, 1.0, 1.0)
	xs2, ts2 = smooth(x, 1.0, 1.0, 2.0)
	for i in range(20):
		print i, x[i], xk[i], xs0[i], xs1[i], xs2[i]
		assert numpy.alltrue( numpy.equal(numpy.less(x, 0.5),
						numpy.less(xk, 0.5)
						)
					)
		assert numpy.alltrue( numpy.equal(numpy.less(x, 0.5),
						numpy.less(xs0, 0.5)
						)
					)
		assert numpy.alltrue( numpy.equal(numpy.less(x, 0.5),
						numpy.less(xs1, 0.5)
						)
					)
		assert numpy.alltrue( numpy.equal(numpy.less(x, 0.5),
						numpy.less(xs2, 0.5)
						)
					)


def test2():
	x = numpy.zeros((500,), numpy.float)
	x[100:] = 1.0
	xs0, ts0 = smooth(x, 0.1, 1.0, 0.01)
	xs1, ts1 = smooth(x, 0.1, 1.0, 1.0)
	xs2, ts2 = smooth(x, 0.1, 1.0, 2.0)
	for i in range(200):
		if i%10 == 5:
			print i, x[i], xs0[i//10], xs1[i//10], xs2[i//10]
		else:
			print i, x[i]

def test3():
	x = numpy.zeros((200,), numpy.float)
	x[100] = 1.0
	xs0, ts0 = smooth(x, 0.1, 0.1, 0)
	xs1, ts1 = smooth(x, 0.1, 0.3, 0)
	xs2, ts2 = smooth(x, 0.1, 0.9, 0)
	xs3, ts3 = smooth(x, 0.1, 1.5, 0)
	for i in range(197):
		o = [ '%d'%i ]
		o.append( '%g' % x[i])
		o.append( '%g' % xs0[i])
		if i%3 == 1:
			o.append('%g' % xs1[i//3])
		else:
			o.append('')

		if i%9 == 5:
			o.append('%g' % xs2[i//9])
		else:
			o.append('')

		if i%15 == 7:
			o.append('%g' % xs3[i//15])
		else:
			o.append('')

		print '\t'.join(o)


if __name__ == '__main__':
	test1()
	test2()
	test3()



#!/usr/bin/env python

"""This simulates re-running feature_vec.py with the -xform or -XFORM flags.
It reads in a feature vector, then transforms it by a matrix multiplication,
and writes it back out.
"""

import os
import sys
import math
from gmisclib import chunkio
from gmisclib import die
import gpkimgclass
from gmisclib import avio
from gmisclib import Num
# from gpk_voicing import feature_vec as FV

PLOT = False

XFORM = None


def fiddle(h):
	o = {}
	for (k, v) in h.items():
		if k.startswith('F_INFO'):
			o['OLD_%s' % k] = v
		elif k.startswith('RMS'):
			pass
		elif k.startswith('TTYPE'):
			pass
		elif k.startswith('TUNITS'):
			pass
		else:
			o[k] = v
	return o


def process(d, Xform):
	fac = chunkio.datachunk(open(Xform, 'r')).read_NumArray()
	o = Num.matrixmultiply(data.d, fac)
	hdr = fiddle( data.hdr )
	hdr['XFORM'] = Xform
	for i in range( fac.shape[1] ):
		hdr['TTYPE%d' % (i+1)] = 'Mixture%d' % i
		d = ','.join(['%g' % q for q in fac[:,i] ])
		hdr['F_INFO%d' % (i+1)] = d
		hdr['RMS%d' % (i+1)] = math.sqrt(Num.average(o[:,i]**2)) 
	gpkimgclass.gpk_img(hdr, o).write(ofile)



if __name__ == '__main__':
	ofile = '-'
	arglist = sys.argv[1:]
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-o':
			ofile = arglist.pop(0)
		elif arg == '-xform':
			fxd = os.environ['OXIVOICE']
			XFORM = '%s/xform.chunk' % fxd
		elif arg == '-Xform':
			XFORM = arglist.pop(0)
		else:
			die.die('Unrecognized flag: %s' % arg)
	assert XFORM, "Nothing to do!"
	fname = arglist[0]
	data = gpkimgclass.read(fname)
	process(data, XFORM)



#!/usr/bin/env python

"""Duration estimator for speech.

Usage: s_slope [flags]
Flags: -f FFF opens file FFF for reading.
	-c XXX  selects column XXX (either a data name or an integer).
	-o FFF  sets the output file.
	-dt #.##   Sets the interval for calculating the output.

It takes a local spectrum,
bins it onto the Bark scale,
converts to perceptual loudness (via **E).
Then, it computes a measure of how far you can go
from each point before the spectrum changes too much.
"""

from gmisclib import Num
from gmisclib import erb_scale

from gpk_voicing import percep_spec as PS
from gpk_voicing import fv_misc as FVM

def feature_vec(data, dt, Dt,
		Nsv=FVM.NSV
		):
	FORMANT_LOW = erb_scale.f_to_erb(200.0)
	FORMANT_HIGH = erb_scale.f_to_erb(4000.0)
	assert Dt > 0.0 and float(Dt)>0.0
	bmin = erb_scale.f_to_erb(100.0)
	bmax = erb_scale.f_to_erb(6000.0)
	# all_ectrs, all_ps, t0 = PS.perceptual_spec(data, dt, Dt, ...
	all_ectrs, all_ps, t0 = PS.block_percep_spec(data, dt, Dt,
							bmin=bmin, bmax=bmax, db=FVM.DB,
							do_mod=0,
							do_dissonance=False,
							PlompBouman=False
							)

	band_indices = [i for (i,ec) in enumerate(all_ectrs) if ec['type']=='band']
	neural = all_ps.take(band_indices, axis=0)
	ectrs = [ec for ec in all_ectrs if ec['type']=='band']
	nband_indices = [i for (i,ec) in enumerate(all_ectrs) if ec['type']!='band']
	nneural = all_ps.take(nband_indices, axis=0)
	nectrs = [ec for ec in all_ectrs if ec['type']!='band']

	assert nneural.shape[1]==neural.shape[1]
	assert neural.shape[1]==all_ps.shape[1]
	assert neural.shape[0]+nneural.shape[0] == all_ps.shape[0]
	assert len(data.shape) == 1
	assert abs(all_ps.shape[1]*Dt-data.shape[0]*dt) < 0.1*data.shape[0]*dt

	neural_now = Num.average(neural, axis=0)
	assert neural_now.shape[0] == neural.shape[1]
	neural_avg = Num.sum(neural_now**2)/Num.sum(neural_now)
	Num.divide(neural, neural_avg, neural)
	Num.divide(neural_now, neural_avg, neural_now)

	assert nneural.shape[0] < nneural.shape[1]
	assert len(nectrs) == nneural.shape[0]
	for (i,e) in enumerate(nectrs):
		assert e['type'] == 'band'

	# print '# neural_avg=', neural_avg
	Vl = 0.04
	w = int(round(Vl/Dt))
	o, descr = FVM.vowel(w, ectrs, neural, neural_now, Nsv,
						formant_low=FORMANT_LOW,
						formant_high=FORMANT_HIGH
						)
	assert len(descr)==len(o), "Descriptor mismatch"
	return (o, descr, t0)




#!/usr/bin/env python

import os
import sys
import math as M
from gmisclib import Num
from gmisclib import die
# from gmisclib import erb_scale
from gmisclib import avio
import gpkavg
import gpkimgclass

# from gpk_voicing import percep_spec
# from gpk_voicing import voice_misc
from gpk_voicing import fv_misc as FVM
# import irregularity
# sys.path.insert(0, '%s/tick1/bin' % os.environ['MRImodel'])

if 'PYLABDISP' in os.environ:
	if os.environ['PYLABDISP'] == 'pylab':
		import pylab
	else:
		import g_pylab as pylab
	PLOT = True
else:
	PLOT = False




# SpecExp = 0.0




def average(x):
	vv = []
	ww = []
	for (v,wt) in x:
		vv.append(v)
		ww.append(wt)
	return gpkavg.avg(vv, ww, 0.0)[0]

def median(x):
	vv = []
	ww = []
	for (v,wt) in x:
		vv.append(v)
		ww.append(wt)
	return gpkavg.avg(vv, ww, 0.499)[0]


def weight(x):
	sum = 0.0
	for (v,wt) in x:
		sum += wt**2
	return M.sqrt(sum)


class matmult_xform(object):
	from gmisclib import chunkio
	def __init__(self, filename):
		self._name = filename
		self.xform = self.chunkio.datachunk(open(filename, 'r')).read_NumArray()

	def name(self):
		return self._name

	def describe_xform(self, descr):
		pass

	def operate(self, o, descr):
		Num.matrixmultiply(o, self.xform, o)
		for i in range(o.shape[1]):
			descr[i] = {'id': 'Mixture%d'% i}



def describe_xform(descr, fac):
	nin, nout = fac.shape
	assert len(descr) == nin
	for i in range(nout):
		norm = M.sqrt(Num.average(fac[:,i]**2))
		nf = fac[:,i]/norm
		tmp = []
		fwt = []
		wwt = []
		edginess = []
		voiciness = []
		burstiness = []
		friciness = []
		vness = []
		for j in range(nin):
			dj = descr[j]
			if dj.has_key('erb'):
				fwt.append( (dj['erb'], abs(nf[j])) )
			if dj.has_key('width'):
				wwt.append( (dj['width'], abs(nf[j])) )
			if 'edge' in dj['type']:
				edginess.append( abs(nf[j]) )
			if 'burst' in dj['type']:
				burstiness.append( abs(nf[j]) )
			if 'fricative' in dj['type']:
				friciness.append( abs(nf[j]) )
			if 'vowel' in dj['type']:
				vness.append( abs(nf[j]) )
			if 'haspitch' in dj['type']:
				voiciness.append( abs(nf[j]) )
			if abs(nf[j]) > 0.1:
				tmp.append('%+.2f*%s' % (nf[j], dj['id']) )
		print '# FV[%d]= %s' % (i, ' '.join(tmp))
		fb = {'voicing': M.sqrt(Num.sum(Num.array(voiciness)**2)),
			'edges': M.sqrt(Num.sum(Num.array(edginess)**2)),
			'burst': M.sqrt(Num.sum(Num.array(burstiness)**2)),
			'fricative': M.sqrt(Num.sum(Num.array(friciness)**2)),
			'vowel': M.sqrt(Num.sum(Num.array(vness)**2)),
			'meanfreq': average(fwt), 'medfreq': median(fwt),
			'wt_freq': weight(fwt),
			'meanwidth': average(wwt), 'medwidth': median(wwt),
			'wt_width': weight(wwt),
			'channel': i, 'description': ' '.join(tmp),
			'ltype': 'fb'
			}
		print avio.concoct(fb)




def run(argv):
	Lf = 1.5
	Elf = 1.5
	Nsv = FVM.NSV
	# Nsv and LF updated from the 5 April 2007 optimization
	# in .../m/ASR/OPT
	DT = 0.01	# Seconds.  Default output sampling interval.
	Scale = None
	arglist = argv[1:]
	arglist0 = arglist
	column = None
	signalfile = None
	outfile = None
	verbose = 0
	DoVoicing = 2
	DoDissonance = True
	import gpk_voicing.fvcurrent as fvm
	extrahdr = {}
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-DT' or arg == '-dt':
			DT = float(arglist.pop(0))
		elif arg == '-f':
			signalfile = arglist.pop(0)
		elif arg == '-c' or arg == '-col':
			tmp = arglist.pop(0)
			try:
				column = int( tmp )
			except ValueError:
				column = tmp
			die.note("signalfile", signalfile)
		elif arg == '-o':
			outfile = arglist.pop(0)
		elif arg == '-v':
			verbose += 1
		elif arg == '-Nsv':
			FVM.NSV = float(arglist.pop(0))
		# elif arg == '-SpecExp':
			# SpecExp = float(arglist.pop(0))
		elif arg == '-sfv':
			import gpk_voicing.fvsimple as fvm
			DoVoicing = 1
		elif arg == '-ssfv':
			import gpk_voicing.fvss as fvm
			DoDissonance = False
			DoVoicing = 1
		elif arg == '-fv20071030':
			import gpk_voicing.fv20071030 as fvm
			DoVoicing = 1
		elif arg == '-fv20080325':
			import gpk_voicing.fv20080325 as fvm
			DoVoicing = 1
		elif arg == '-fv200807jyuan':
			import gpk_voicing.fv200807jyuan as fvm
			DoVoicing = 1
			DoDissonance = True
		elif arg == '-fv200811align':
			import gpk_voicing.fv200811align as fvm
			DoVoicing = 2
			DoDissonance = True
		elif arg == '-fv200812align':
			import gpk_voicing.fv200812align as fvm
			DoVoicing = 2
			DoDissonance = True
			Lf = 1.0
			Elf = 1.0
		elif arg == '-fv200908opt':
			import gpk_voicing.fv200908opt as fvm
			DoVoicing = 1
			DoDissonance = False
		elif arg.startswith('-fv200909opt'):
			import gpk_voicing.fv200909opt as fvm
			DoVoicing = 1
			DoDissonance = False
			Nsv, Lf, Elf = fvm.Scale.get_many('nsv', 'lf,lf', 'lf,elf')
			if arg.endswith('S'):
				Scale = fvm.Scale
		elif arg == '-LF':
			Lf = float(arglist.pop(0))
		elif arg == '-ELF':
			Elf = float(arglist.pop(0))
		elif arg == '-novoicing':
			DoVoicing = 0
		elif arg == '-Xform':
			Scale = matmult_xform(arglist.pop(0))
		elif arg == '-write':
			extrahdr = dict( [q.strip().split('=', 1)
					for q in arglist.pop(0).split(';') ]
					)
		else:
			die.info("Unrecognized flag: %s" % arg)
			print __doc__
			die.exit(1)
	if arglist and signalfile is None:
		signalfile = arglist.pop(0)
	if arglist and outfile is None:
		outfile = arglist.pop(0)
	elif outfile is None:
		outfile = 'fv.dat'
	if signalfile is None:
		die.info("No signal file specified.")
		print __doc__
		die.exit(1)
	if arglist:
		die.info('Extra arguments!')
		print __doc__
		die.exit(1)
	signal = gpkimgclass.read(signalfile)
	if column is None and signal.d.shape[1]==1:
		column = 0
	elif column is None:
		die.die("There are %d channels in %s, so you must specify one with '-c'." % (signal.d.shape[1], signalfile))
	try:
		signal.column(column)
	except KeyError:
		die.info("File has %d columns" % signal.n[1])
		die.die("Bad column: %s" % str(column))

	o, descr, DTx, tshift = fvm.feature_vec(signal.column(column), signal.dt(), DT,
					Nsv=Nsv, LF=Lf, ELF=Elf,
					do_voicing=DoVoicing,
					do_dissonance=DoDissonance
					)
	o = Num.transpose(o)
	hdr = signal.hdr.copy()
	if Scale is not None:
		Scale.operate(o, descr)
		hdr['XFORM'] = Scale.name()
		if verbose:
			Scale.describe_xform(descr)

	hdr['program'] = argv[0]
	hdr['ARGV'] = arglist0
	hdr['input_file'] = signalfile
	hdr['column'] = column
	hdr['CDELT2'] = DTx
	hdr['CRPIX2'] = 1
	hdr['CRVAL2'] = signal.start() + tshift
	hdr['CDELT1'] = 1
	hdr['BITPIX'] = -32
	hdr['DataSamplingFreq'] = 1.0/signal.dt()
	hdr['Nsv'] = FVM.NSV
	hdr['LF'] = Lf
	hdr['ELF'] = Elf
	hdr.update( extrahdr )

	if PLOT:
		pylab.matshow(Num.transpose(o))
		pylab.show()
	assert o.shape[1] == len(descr), "Length mismatch data=%s descr=%d" % (
						o.shape, len(descr)
						)
	for (i,d) in enumerate(descr):
		hdr['TTYPE%d' % (i+1)] = d['id']
		hdr['F_INFO%d' % (i+1)] = avio.concoct(d)
		hdr['RMS%d' % (i+1)] = M.sqrt(Num.average(o[:,i]**2)) 
	gpkimgclass.gpk_img(hdr, o).write(outfile)



if __name__ == '__main__':
	# try:
		# import psyco
		# psyco.full()
	# except ImportError:
		# pass

	run(sys.argv)



import math
import numpy
from gpk_voicing import voice_misc
from gpk_voicing import power




def one_pk_pk(t, data, dt, Pkwdt):
	"""Measure the peak-to-peak amplitude near time t.
	@param t: the time to measure near (units of seconds)
	@type t: float (seconds).
	@param data: the time-series.
	@type data: numpy.ndarray
	@param dt: the sampling interval for data
	@type dt: float (seconds).
	@param Pkwdt: Half-width of region over which to compute
		the peak-to-peak amplitude.   The low end of the
		frequency response will be on the order of 1/Pkwdt.
	@type Pkwdt: float (seconds).
	@return: scalar peak-to-peak amplitude
	@rtype: float
	"""
	s, e = voice_misc.start_end(t, dt, int(round(Pkwdt/dt)), data.shape[0])
	d = data[s:e]
	return d[ numpy.argmax( d ) ] - d[ numpy.argmin(d) ]


def compute_intensity(data, dt, DT, DR, HPF, DTSmooth):
	"""Measure the time-series of RMS of a data array data.
	DT is the sampling interval for the resulting RMS time series."""
	if HPF is not None:
		data = voice_misc.hipass_sym_butterworth(data, HPF*dt)
	if DR == 'rms':
		pwr, t0 = power.smooth(data**2, dt, DT, DTSmooth)
		rms = numpy.sqrt(pwr)
	elif DR == 'avd':
		rms,t0 = power.smooth(numpy.absolute(data), dt, DT, DTSmooth)
	elif DR == 'pkpk':
		ns = int(math.floor(data.shape[0]*(dt/DT)))
		# print 'ns=', dt, data.shape[0], DT, dt*data.shape[0]/DT
		pkpk = numpy.zeros((ns,))
		for t in range(ns):
			pkpk[t] = one_pk_pk(t*DT, data, dt, math.hypot(DTSmooth, DT))
		rms,t0 = power.smooth(pkpk, DT, DT, DTSmooth)
	else:
		raise ValueError, 'Bad DR=%s' % DR
	return (rms,t0)





#! python

"""A little program to check that voice_misc._fftwork() is correct."""


from gpk_voicing import voice_misc
import numpy
import pylab
import time


def plot(start, end):
	x = []
	y = []
	z = []
	ws = 0.0
	ts = 0.0
	for n in range(start, end):
		x.append(n)
		w = voice_misc._fftwork_real(n)
		# w = voice_misc._fftwork(n)
		y.append(w)
		q = numpy.zeros((n,))
		k = 0
		t0 = time.clock()
		while time.clock() < t0 + 0.1:
			# numpy.fft.fft(q)
			s = numpy.fft.rfft(q)
			numpy.fft.irfft(s)
			k += 1
		t1 = time.clock()
		dt = (t1-t0)/k
		z.append(dt)
		ts += dt
		ws += w
	wscaled = [w*(ts/ws) for w in y]
	pylab.plot(x, wscaled, 'g')
	pylab.plot(x, z, 'r+')
	pylab.show()


if __name__ == '__main__':
	plot(127, 257)




import re
import math

from gmisclib import die
from gmisclib import Num
from gmisclib import gpkmisc
from gmisclib import g_encode
from gmisclib import erb_scale
from gmisclib import Numeric_gpk as NG
from gpk_voicing import voice_misc
try:
	from gpk_voicing import pseudo_dur as PD
except ImportError:
	PD = None

pylab = None
# import pylab

DB = math.sqrt(0.5)
NSV = 0.75
# Dt = 0.005
C = 10.0


class SillyWidthException(ValueError):
	def __init__(self, *s):
		ValueError.__init__(self, *s)


# def edgepair_win(n):
	# e = Num.zeros((2*n,), Num.Float)
	# e[:n] = -1.0
	# e[n:] = 1.0
	# s = Num.ones((2*n), Num.Float)
	# return (e, s)


def edgepair_win(n):
	if not n > 0:
		raise SillyWidthException, n
	tmp0 = voice_misc.window(2*n, 0)
	tmp1 = voice_misc.window(2*n, 1)
	norm = gpkmisc.N_maximum(tmp0)
	return (tmp1/norm, tmp0/norm)


# def win(n):
	# tmp = Num.ones((n,), Num.Float)
	# return tmp

def win(n):
	if not n > 0:
		raise SillyWidthException, n
	tmp = voice_misc.window(n, 0)
	return tmp/gpkmisc.N_maximum(tmp)


def burst_win(n, i):
	if not n > 0:
		raise SillyWidthException, n
	tmp = Num.zeros((n,), Num.Float)
	tmp[i] = 1.0
	neg = win(n)
	nneg = neg/Num.sum(neg)
	return (tmp - nneg, nneg)


def entropy(w):
	assert Num.alltrue( Num.greater(w, 0.0) )
	p = w/Num.sum(w)
	return -Num.sum(p*Num.log(p))

def convolve(signal, kernel):
	"""Returns something the length of the signal, by zero padding."""
	if signal.shape[0] > kernel.shape[0]:
		return Num.convolve(signal, kernel, 1)
	m = 1+kernel.shape[0]-signal.shape[0]
	tmp = Num.concatenate((signal, Num.zeros((m,), Num.Float)))
	die.info('Narrow signal: padding from %d to %d' % (signal.shape[0], m))
	return Num.convolve(tmp, kernel, 1)[:signal.shape[0]]


def vowel(width, ectrs, neural, neural_now, Nsv,
		formant_low=None, formant_high=None):
	VFAC = 0.75
	cs = win(width)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)/Num.sum(cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	# The second term within the hypot() should always be Nsv*one.

	# pylab.plot(norm)
	# pylab.figure()
	o = []
	descr = []
	for (i, e) in enumerate(ectrs):
		if e['type']=='band' and formant_low < e['erb'] < formant_high:
			tmp = (VFAC/css)*convolve(neural[i], cs)/norm
			# pylab.plot(tmp)
			o.append(tmp)
			# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
			dtmp = e.copy()
			dtmp['type'] = 'vowel'
			dtmp['width'] = width
			dtmp['Kentropy'] = entropy(cs)
			dtmp['Fentropy'] = 0.0
			dtmp['id'] = 'vowel:%d:%.1f' % (width, e['erb'])
			dtmp['t_symmetry'] = 1
			dtmp['a_scaling'] = 1
			# pylab.title('vowel %s' % dtmp['id'])
			# pylab.figure()
			descr.append( dtmp )
	# pylab.show()
	return (o, descr)


def spectral_entropy(width, ectrs, neural, neural_now, Nsv):
	"""This is inspired (but only loosely) by
	"Robust Entropy-based Endpoint Detection for Speech Recognition
	in Noisy Environments." by Jia-lin Shen and Jeih-weih Hung
	and Lin-shan Lee, http://www.ee.columbia.edu/~dpwe/papers/ShenHL98-endpoint.pdf
	International Conference on Spoken Language Processing, 1998.
	"""
	assert Nsv > 0.0
	SEF = 1.0
	EPS = 1e-6
	cs = win(width)
	assert Num.alltrue( Num.greater(cs, 0.0) )

	nnt = []
	nns = Num.zeros(neural.shape[1], Num.Float)
	for (i, e) in enumerate(ectrs):
		if e['type']=='band':
			assert Num.alltrue( Num.greater_equal(neural[i], 0.0) )
			tmp = convolve(neural[i], cs)
			nnt.append(tmp)
			Num.add(nns, tmp, nns)

	np = len(nnt)
	ent_sum = Num.zeros(neural.shape[1], Num.Float) + math.log(4.0)
	for tmp in nnt:
		Num.divide(tmp, nns, tmp)
		Num.add(tmp, EPS*nns, tmp)
		etmp = tmp*Num.log(tmp)
		Num.add(ent_sum, etmp, ent_sum)
	Num.multiply(ent_sum, -SEF, ent_sum)
	# pylab.title('spectral_entropy')
	# pylab.show()
	assert np > 0
	dtmp = {'type': 'spectral_entropy',
		'width': width,
		'Kentropy': entropy(cs),
		'Fentropy': math.log(np),
		'id': 'Sentropy:%d' % width,
		't_symmetry': 1,
		'a_scaling': 0
		}
	return ([ent_sum], [dtmp])


def space_time_entropy(width, ectrs, neural, neural_now, Nsv):
	assert Nsv > 0.0
	SEF = 1.0
	EPS = 1e-6
	cs = win(width)

	nnt = []
	nns = Num.zeros(neural.shape[1], Num.Float)
	for (i, e) in enumerate(ectrs):
		if e['type']=='band':
			assert Num.alltrue( Num.greater_equal(neural[i], 0.0) )
			nnt.append(neural[i])
			Num.add(nns, neural[i], nns)

	np = len(nnt)
	nns = convolve(nns, cs)
	ent_sum = Num.zeros(neural.shape[1], Num.Float) + math.log(3.0*np)
	for tmp in nnt:
		Num.divide(tmp, nns, tmp)
		Num.add(tmp, EPS*nns, tmp)
		etmp = convolve(tmp*Num.log(tmp), cs)
		Num.add(ent_sum, etmp, ent_sum)
	Num.multiply(ent_sum, -SEF, ent_sum)
	# pylab.title('spectral_entropy')
	# pylab.show()
	assert np > 0
	dtmp = {'type': 'space_time_entropy',
		'width': width,
		'Kentropy': entropy(cs),
		'Fentropy': math.log(np),
		'id': 'STentropy:%d' % width,
		't_symmetry': 1,
		'a_scaling': 0
		}
	return ([ent_sum], [dtmp])



def haspitch(width, ectrs, neural, neural_now, Nsv):
	HFAC = {'high': 20.0, 'all': 10.0, 1: 10.0}
	cs = win(width)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)/Num.sum(cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	o = []
	descr = []
	for (i, e) in enumerate(ectrs):
		if e['type'].startswith('haspitch'):
			tmp = (HFAC[e['variant']]/css)*convolve(neural[i], cs)/norm
			# pylab.figure()
			# pylab.title('e=%s' % e['id'])
			# pylab.plot(neural[i])
			# pylab.plot(tmp)
			# pylab.plot(norm)
			o.append(tmp)
			# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
			dtmp = e.copy()
			dtmp['id'] = '%s:%d' % (dtmp['id'], width)
			dtmp['Kentropy'] = entropy(cs)
			assert 'Fentropy' in dtmp
			dtmp['width'] = width
			dtmp['t_symmetry'] = 1
			dtmp['a_scaling'] = 1
			descr.append( dtmp )
	# pylab.show()
	return (o, descr)



def peakiness(width, ectrs, neural, neural_now, Nsv):
	HFAC = 10
	cs = win(width)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)/Num.sum(cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	o = []
	descr = []
	for (i, e) in enumerate(ectrs):
		if e['type'].startswith('peakalign'):
			tmp = (HFAC/css)*convolve(neural[i], cs)/norm
			# pylab.figure()
			# pylab.title('e=%s' % e['id'])
			# pylab.plot(neural[i])
			# pylab.plot(tmp)
			# pylab.plot(norm)
			o.append(tmp)
			# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
			dtmp = e.copy()
			dtmp['id'] = '%s:%d' % (dtmp['id'], width)
			dtmp['Kentropy'] = entropy(cs)
			assert 'Fentropy' in dtmp
			dtmp['width'] = width
			dtmp['t_symmetry'] = 1
			dtmp['a_scaling'] = 1
			descr.append( dtmp )
	# pylab.show()
	return (o, descr)


def pdur(ectrs, neural, neural_now, Dt, Cfac=1.0):
	if PD is None:	# If pseuduration is not available.
		return ([], [])
	TYPICAL = 0.100
	c = C * Cfac
	norm = Num.sum(neural_now**2)/Num.sum(neural_now)
	no = neural_now.shape[0]
	out = Num.zeros((no,), Num.Float)
	for t in range(no):
		plm, pcm = PD.pdur_guts(neural.transpose(), t, -1, Dt, c/norm**2)
		plp, pcp = PD.pdur_guts(neural.transpose(), t, 1, Dt, c/norm**2)
		out[t] = plm + plp
	if pylab:
		pylab.figure()
		pylab.title('pdur')
		pylab.plot(out)
		pylab.show()
	nbands = 0
	for e in ectrs:
		if e['type']=='band':
			nbands += 1
	assert nbands > 0
	typical = NG.N_median(out)
	descr = [ {'id': 'pseudoduration:%f' % c,
		'Kentropy': NG.N_median(Num.log(out)),
		'Fentropy': math.log(nbands),
		'width': typical,
		't_symmetry': 1, 'a_scaling': 0
		} ]
	return ([Num.log(out/TYPICAL)], descr)


def roughness(width, ectrs, neural, neural_now, Nsv):
	HFAC = {1: 4.0}
	cs = win(width)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)/Num.sum(cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	o = []
	descr = []
	for (i, e) in enumerate(ectrs):
		if e['type'].startswith('roughness'):
			tmp = (HFAC[e['variant']]/css)*convolve(neural[i], cs)/norm
			# pylab.figure()
			# pylab.title('e=%s' % e['id'])
			# pylab.plot(neural[i])
			# pylab.plot(tmp)
			# pylab.plot(norm)
			o.append(tmp)
			# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
			dtmp = e.copy()
			dtmp['id'] = '%s:%d' % (dtmp['id'], width)
			dtmp['Kentropy'] = entropy(cs)
			assert 'Fentropy' in dtmp
			dtmp['width'] = width
			dtmp['t_symmetry'] = 1
			dtmp['a_scaling'] = 1
			descr.append( dtmp )
	# pylab.show()
	return (o, descr)

dissonance = roughness	# obsolete


def vowel_edge(width, ectrs, neural, neural_now, Nsv, do_abs=False,
			FORMANT_LOW=None, FORMANT_HIGH=None):
	VEfac = 0.7
	ce, cs = edgepair_win(width)
	nns = convolve(neural_now, cs)/Num.sum(cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	# pylab.plot(norm)
	# pylab.figure()
	o = []
	descr = []
	for (i, e) in enumerate(ectrs):
		if e['type']=='band' and FORMANT_LOW < e['erb'] < FORMANT_HIGH:
			tmp = VEfac * convolve(neural[i], ce)/norm
			# pylab.plot(tmp)
			# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
			dtmp = e.copy()
			dtmp['width'] = 2*width
			dtmp['Kentropy'] = entropy(cs)
			dtmp['Fentropy'] = 0.0
			dtmp['t_symmetry'] = -1
			dtmp['a_scaling'] = 1
			if do_abs:
				o.append( Num.absolute(tmp) )
				dtmp['type'] = 'vowel |edge|'
				dtmp['id'] = 'vowel |edge|:%d:%.1f' % (width, e['erb'])
				descr.append( dtmp )
			else:
				o.append( tmp )
				dtmp['type'] = 'vowel edge'
				dtmp['id'] = 'vowel edge:%d:%.1f' % (width, e['erb'])
				descr.append( dtmp )
	# pylab.show()
	return (o, descr)


def fricative(width, ectrs, neural, neural_now, Nsv):
	CSSE = 0.7
	N = neural[0].shape[0]
	cs = win(width)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)/css
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	# pylab.plot(norm)
	# pylab.figure()
	o = []
	descr = []
	for (rs, re, fac) in [(70.0, 700.0, 0.6), (700.0, 1500.0, 0.6),
				(1500.0, 2500.0, 0.9), (2500.0, 4000.0, 1.1),
				(4000.0, 6000.0, 1.3)
				]:
		tsum = Num.zeros((N,), Num.Float)
		elow = erb_scale.f_to_erb(rs)
		ehigh = erb_scale.f_to_erb(re)
		included = []
		nq = 0
		for (i, e) in enumerate(ectrs):
			if e['type']=='band' and elow < e['erb'] < ehigh:
				Num.add(tsum, neural[i], tsum)
				included.append(e)
				nq += 1
		tmp = (fac/css**CSSE) * convolve(tsum, cs)/(norm*nq)
		# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
		# pylab.plot(tmp + len(o))
		o.append( tmp )
		descr.append( {'type': 'fricative', 'width': width,
				'erbs': [e['erb'] for e in included],
				'fcs' : [e['fc'] for e in included],
				'Kentropy': entropy(cs),
				'Fentropy': -math.log(len(included)),
				'id': 'fricative:%d:%.1f-%.1f' % (width, elow, ehigh),
				't_symmetry': 1, 'a_scaling': 1
				}
				)
	# pylab.show()
	return (o, descr)


def prominence(ectrs, neural, neural_now, Nsv):
	WIDTH = 20
	N = neural[0].shape[0]
	ce, cs = edgepair_win(WIDTH)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	# pylab.plot(norm)
	# pylab.figure()
	o = []
	descr = []
	included = []
	for (rs, re, fac) in [ (70.0, 600.0, 0.22), (600.0, 2000.0, 0.25),
				(2000.0, 3000.0, 0.32), (3000.0, 6000.0, 0.35) ]:
		tsum = Num.zeros((N,), Num.Float)
		elow = erb_scale.f_to_erb(rs)
		ehigh = erb_scale.f_to_erb(re)
		for (i, e) in enumerate(ectrs):
			if e['type']=='band' and elow < e['erb'] < ehigh:
				Num.add(tsum, neural[i], tsum)
				included.append(e)
		tmp = (fac/css) * convolve(tsum, ce)/norm
		# pylab.plot(tmp + len(o))
		# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
		o.append( tmp )
		dtmp = {'type': 'prominence', 'width': WIDTH,
				'erbs': [e['erb'] for e in included],
				'fcs' : [e['fc'] for e in included],
				'Kentropy': entropy(cs),
				'Fentropy': -math.log(len(included)),
				'id': 'prominence:%d:%.1f-%.1f' % (WIDTH, elow, ehigh),
				't_symmetry': 1, 'a_scaling': 1
				}
		descr.append( dtmp )
	# pylab.show()
	return (o, descr)




def fricative_edge(width, ectrs, neural, neural_now, Nsv, do_abs=False):
	CSSE = 1.6
	N = neural[0].shape[0]
	ce, cs = edgepair_win(width)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)/css
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	# pylab.plot(norm)
	# pylab.figure()
	o = []
	descr = []
	for (rs, re, fac) in [(70.0, 700.0, 7), (700.0, 1500.0, 7),
				(1500.0, 2500.0, 7), (2500.0, 4000.0, 7),
				(4000.0, 6000.0, 7)
				]:
		tsum = Num.zeros((N,), Num.Float)
		elow = erb_scale.f_to_erb(rs)
		ehigh = erb_scale.f_to_erb(re)
		included = []
		nq = 0
		for (i, e) in enumerate(ectrs):
			if e['type']=='band' and elow < e['erb'] < ehigh:
				Num.add(tsum, neural[i], tsum)
				included.append(e)
				nq += 1
		tmp = (fac/css**CSSE) * convolve(tsum, ce)/(norm*nq)
		# pylab.plot(tmp)
		# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])

		dtmp = {'width': width, 
				'erbs': [e['erb'] for e in included],
				'fcs' : [e['fc'] for e in included],
				'Kentropy': entropy(cs),
				'Fentropy': -math.log(len(included)),
				'a_scaling': 1
				}

		if do_abs:
			o.append( Num.absolute(tmp) )
			dtmp['type'] = 'fricative |edge|'
			dtmp['t_symmetry'] = 1
		else:
			o.append( tmp )
			dtmp['type'] = 'fricative edge'
			dtmp['t_symmetry'] = -1
		dtmp['id'] = '%s:%d:%.1f-%.1f' % (dtmp['type'], width, elow, ehigh)
		descr.append( dtmp )
	# pylab.show()
	return (o, descr)


def burst(ectrs, neural, neural_now, Nsv):
	BKGWIDTH = 8
	WIDTH = 1
	N = neural[0].shape[0]
	ce,cs = burst_win(BKGWIDTH, 2)
	nns = convolve(neural_now, cs)/Num.sum(cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	# pylab.plot(norm)
	# pylab.figure()
	o = []
	descr = []
	included = []
	# Burst bandwidth needs to be large enough so that the
	# burst can fit into a 10ms slot.
	for (rs, re, fac) in [ (150.0, 500.0, 1.0), (500.0, 1200.0, 1.0),
				(1200.0, 2800.0, 1.0), (2800.0, 6000.0, 1.0) ]:
		tsum = Num.zeros((N,), Num.Float)
		elow = erb_scale.f_to_erb(rs)
		ehigh = erb_scale.f_to_erb(re)
		for (i, e) in enumerate(ectrs):
			if e['type']=='band' and elow < e['erb'] < ehigh:
				Num.add(tsum, neural[i], tsum)
				included.append( e )
		tss = fac * convolve(tsum, ce)/norm
		# pylab.plot(tmp + len(o))
		# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
		o.append( tss )
		descr.append( {'type': 'burst', 'width': WIDTH,
				'bkgwidth': BKGWIDTH,
				'erbs': [e['erb'] for e in included],
				'fcs' : [e['fc'] for e in included],
				'Kentropy': entropy(cs),
				'Fentropy': -math.log(len(included)),
				'id': 'burst:%d:%.1f-%.1f'
					% (WIDTH, elow, ehigh),
				't_symmetry': 1, 'a_scaling': 1
				}
				)
	# pylab.show()
	return (o, descr)



class scale_xform(object):
	def __init__(self, opt_string, name='???', prefix='scale'):
		self.prefix = prefix
		self._name = name
		self.sc = []
		for tmp in opt_string.split('\n'):
			if tmp=='' or tmp.startswith('#'):
				continue
			v, k = tmp.split()
			self.sc.append( (re.compile(k), float(v)) )
		self._e = g_encode.encoder(notallowed=' %,')

	def get(self, k):
		for (pat,v) in self.sc:
			if pat.match(k):
				return v
		raise KeyError, "scale_xform.get(%s): no matching pattern." % k

	def get_many(self, *k):
		for q in k:
			yield self.get(q)

	def name(self):
		return self._name

	def operate(self, o, descr):
		assert o.shape[1] == len(descr)
		for (i,d) in enumerate(descr):
			k = '%s,%s' % (self.prefix, self._e.fwd(d['id']))
			scalefac = math.exp(self.get(k))
			Num.multiply(o[:,i], scalefac, o[:,i])
			d['id'] = d['id'] + '(scaled)'
			d['scalefac'] = scalefac

	def describe_xform(self):
		pass



#!/usr/bin/env python

"""Perceptual spectrum for speech."""

import sys
import math as M
from gmisclib import erb_scale
from gmisclib import gpkmisc
from gmisclib import die

import numpy

# import wavio
# import gpkimgclass

pylab = None

import gpk_voicing.voice_misc as VM
from gpk_voicing import power
from gpk_voicing import percep_spec_extras as PS
from gpk_voicing import gammatone

CBmax = erb_scale.f_to_erb(8000.0)
CBmin = erb_scale.f_to_erb(50.0)
BBSZ = 0.5
E = 0.333
Neural_Tick = 1e-4	# The best possible time resolution of the auditory
			# system.  (E.g. from binaural direction measurements.)
DISS_BWF = 0.25
MAX_DISS_FREQ = 250.0
HP_DISS_FREQ = 30.0
BLOCK_EDGE = 0.3	# Seconds

# def SafeExp(x):
	# return Num.exp(Num.maximum(-200.0, x))


def tau_lp(fc):
	"""This is the cutoff frequency of the modulation transfer function
	in the ear, as a function of frequency.
	From R. Plomp and M. A. Bouman, JASA 31(6), page 749ff, June 1959
		'Relation of hearing threshold and duration of tone pulses'
		"""
	Tau200 = 0.4
	Tau10k = 0.13
	tmp = Tau200 + (M.log(fc/200.0)/M.log(1e4/200.0))*(Tau10k-Tau200)
	assert tmp>0
	return tmp


def cochlear_filter(fofc):
	"""Transfer function taken from "Improved Audio Coding Using a
	Psychoacoustic Model Based on a Cochlear Filter Bank
	IEEE Transactions of Speech and Audio Processing, 10(7) October 2002,
	Pages 495-503.  au=Frank Baumgarte.
	"""
	Q = 4.0
	Slp = 25.0/(20.0*M.log10(1.2))
	Shp = 8.0/(20.0*M.log10(1.2))
	fofc_Shp = fofc**Shp
	return (fofc_Shp) / ( (1+fofc**Slp) * (1 + (1j/Q)*(fofc**(Shp/2))
				- fofc_Shp)
				)


def threshold(f):
	"""IN pressure amplitude.
	Crudely taken from Handbook of Perception vol 4: hearing,
	E.C.Carterette and M.P.Friedman, editors,
	Academic Press 1978, isbn 0-12-161904-4.
	Curve near 70db used."""
	return M.sqrt((200.0/f)**4 + 1 + (f/8000.0)**16)


def accum_abs_sub(a, b, o):
	"""This does a little computation in a block-wise fashion
	to keep all the data witin the processor's cache.
	It computes sum( abs(a-b) ).
	"""
	BKS = 6000
	n = o.shape[0]
	assert a.shape == (n,)
	assert b.shape == (n,)
	assert o.shape == (n,)
	tmp = numpy.zeros((BKS,), numpy.float)
	for i in range(0, n, BKS):
		e = min(i+BKS, n)
		tmpi = tmp[:e-i]
		ai = a[i:e]
		bi = b[i:e]
		oi = o[i:e]
		numpy.subtract(ai, bi, tmpi)
		numpy.absolute(tmpi, tmpi)
		numpy.add(oi, tmpi, oi)
	return o


def list_accum_abs_sub_dly(tk, dly):
	lng = tk[0][dly:].shape[0]
	o = numpy.zeros((lng,), numpy.float)
	for tkd in tk:
		a = tkd[:-dly]
		b = tkd[dly:]
		accum_abs_sub(a, b, o)
	return o



def process_voicing(tk, tick, Dt):
	"""@param Dt: the sampling rate of the output vectors.
	@type Dt: C{float}
	@param tick: the sampling rate of the vectors in C{tk}.
	@type tick: C{float}
	@param tk: a list of vectors, each of which represents the neural firing
		of a particular point in the cochlea.
	@type tk: [ numpy.ndarray, ...]
	@return: (avg, min) where avg-min is the voicing estimator.
		avg: C{float}, min: C{float}
	"""
	LOWF = 50.0
	HIGHF = 400.0
	lowf = int(round(1.0/(tick*LOWF)))
	highf = int(round(1.0/(tick*HIGHF)))
	tmps = None
	tmpmin = None
	## tmpss = None
	tmpn = 0
	for dly in range(highf, lowf):
		# dhl = dly//2
		# dhr = dly - dhl
		# dlytmp = list_accum_abs_sub_dly(tk, dly)
		dlytmp = PS.list_accum_abs_sub_dly(tk, dly)
		assert len(dlytmp.shape) == 1
		# al = Num.zeros((dhl,), Num.Float)
		# ar = Num.zeros((dhr,), Num.Float)
		# dlytmp = Num.concatenate((al, dlytmp, ar))
		tmp2, t0 = power.smooth(dlytmp, tick, Dt, extra=1.0/(LOWF*M.sqrt(12.0)))

		if tmps is None:
			tmpn = 1
			tmps = tmp2
			tmpmin = numpy.array(tmp2, copy=True)
		else:
			tmpn += 1
			numpy.add(tmps, tmp2, tmps)
			numpy.minimum(tmpmin, tmp2, tmpmin)
	return (tmps, tmpn, tmpmin)


class roughness_c(object):
	"""This should be based on Hutchinson and KNopoff 1978
	Kameoka an Kuriygawa 1969a; Viemeister 1988;
	Aures 1985; Plomp and Levelt 1965.
	But it isn't yet.

	Hutchinson W. and Knopoff, L. 1978 The acoustic componenet of
	Western consonance.  Interface 7, 1-29.
	
	Kameoka and Kuriygawa 1969.  Consonance Theory: Part 1.
	J. Acoustical Soc. America. 45 1451-1458
	(and 1459-1469 for part II).

	Viemeister, N. F. 1977 Temporal factors in audition: A system
	analysis approach.   In E. F. Evans and J. P. Wilson (eds)
	Psychophysics and physiology of hearing (pp. 419-429)
	London: Academic Press

	Aures, W. 1985 Ein Berechnun gsverfahren der Rauhigkeit
	[A roughness calculation method] Acustica 58 268-281

	Plomp R. and Levelt W. 1965 Tonal consonance and critical
	bandwidth.  JASA 38, 548-560.
	"""
	def __init__(self):
		self.y = None
		self.n = 0
		self.dt = None
	
	def add(self, y, dt, diss_bw):
		if self.dt is None:
			assert dt > 0
			self.dt = dt
		else:
			assert abs(dt-self.dt) < 0.0001*self.dt
		tmp = VM.lowpass_sym_butterworth(y, diss_bw*dt)
		tmp = VM.hipass_first_order(tmp, HP_DISS_FREQ*dt)
		tmp = numpy.absolute(tmp)
		# if pylab:
			# print 'roughness_c.add', numpy.average(tmp), diss_bw
			# pylab.plot(tmp)
		if self.y is None:
			self.y = tmp
		else:
			numpy.add(self.y, tmp, self.y)
		self.n += 1

	def get(self, Dt):
		# if pylab:
			# pylab.figure()
		yfilt, t0 = power.smooth(self.y, self.dt, Dt)
		# if pylab:
			# pylab.plot(yfilt)
			# pylab.show()
		return yfilt*(0.1/self.n)


class peakalign_c(object):
	def __init__(self):
		self.y = None
		self.shifts = []
		self.dbg = []
		self.n = 0
		self.dt = None

	EDGE = 700
	
	def add(self, y, dt, fc):
		if self.dt is None:
			assert dt > 0
			self.dt = dt
		else:
			assert abs(dt-self.dt) < 0.0001*self.dt
		FF = -3.2
		sh = int(round(FF/(dt*fc))) + self.EDGE
		if self.y is None:
			self.y = numpy.zeros((y.shape[0]+2*self.EDGE,))
		numpy.add(self.y[sh:sh+y.shape[0]], y, self.y[sh:sh+y.shape[0]])
		if pylab and fc>400:
			self.dbg.append(y)
			self.shifts.append(sh)
		self.n += 1


	def get(self, Dt):
		if pylab:
			for (i, (y,sh)) in enumerate(zip(self.dbg, self.shifts)):
				xpl = []
				ypl = []
				for dly in range(-100,100):
					ypl.append(numpy.sum(y*self.y[sh+dly:sh+y.shape[0]+dly]))
					xpl.append(dly)
				pylab.plot(xpl, [q/max(ypl) for q in ypl], linestyle='-',
						color=(float(i)/float(self.n), 0.5, float(self.n-i)/float(self.n)))
		if pylab:
			pylab.show()
		cfilt, t0 = power.smooth(self.y*self.y, self.dt, Dt)
		return cfilt/self.n**2


def block_percep_spec(data, dt, Dt, **kwargs):
	"""This computes the perceptual spectrum in blocks and glues them
	together.   It's useful when the data is too large to fit in memory.
	"""
	if kwargs.get('blocksize', None) is None:
		return perceptual_spec(data, dt, Dt, **kwargs)
	assert BLOCK_EDGE >= Dt
	blocksize = Dt * int(round(kwargs['blocksize']/Dt))
	del kwargs['blocksize']
	assert blocksize > BLOCK_EDGE
	m = data.shape[0]
	ntau = int(M.ceil(m*dt/blocksize))
	nout = int(M.ceil(m*dt/Dt))
	pad = int(round(Dt*round(BLOCK_EDGE/Dt)/dt))

	neural = None
	tshift = None
	bctrs = None
	S0 = 0.0
	e = 0
	for i in range(ntau):
		s0 = int(round((m*i)/float(ntau)))
		se = int(round((m*(i+1))/float(ntau)))
		ps0 = max(0, s0-pad)
		pse = min(m, se+pad)
		tbc, x, ttshift = perceptual_spec(data[ps0:pse], dt, Dt, **kwargs)
		if tshift is None:
			tshift = ttshift
		if bctrs is None:
			bctrs = tbc
		else:
			assert tbc == bctrs
		if neural is None:
			neural = numpy.zeros((x.shape[0], nout+1), numpy.float)

		tau0 = int(M.floor((s0-ps0)*(dt/Dt)))
		nS = min((se-s0)*(dt/Dt), x.shape[1]-tau0)
		# print 'dt=', dt, "Dt=", Dt
		# print 'nS=', nS, "min(", (se-s0)*(dt/Dt), x.shape[0]-tau0, ")"
		inS = int(M.ceil(nS))
		iS0 = int(M.floor(S0))
		assert iS0+inS <= nout, "S0=%.1f nS=%.1f S0+nS=%.1f nout=%d" % (S0, nS, S0+nS, nout)
		# print 'neural', neural.shape, iS0, iS0+inS, 'x=', x.shape, tau0, tau0+inS
		neural[:, iS0:iS0+inS] = x[:, tau0:tau0+inS]
		# print 'S0=', S0, '->', S0+nS
		S0 += nS
		e = iS0+inS
		# print 'e=', e
	assert nout-2 <= e <= nout+1, "e=%d nout=%d" % (e, nout)
	return (bctrs, neural[:,:e], tshift)


class fft_filterbank(object):
	def __init__(self, data, dt, Dt):
		self.d = VM.dataprep_flat_real(data, dt, Dt)
		self.ss = numpy.fft.real_fft( self.d )
		self.f = numpy.absolute(VM.real_fft_freq(self.ss.shape[0], d=dt))

	def filter(self, fc):
		xferfcn = cochlear_filter(self.f/fc)
		xferfcn[0] = xferfcn[0].real
		xferfcn[-1] = xferfcn[-1].real
		return numpy.fft.inverse_real_fft( self.ss * xferfcn )


class gammatone_filterbank(object):
	def __init__(self, data, dt, Dt):
		self.dt = dt
		self.data = data

	def filter(self, fc):
		return gammatone.gammatone(self.data, self.dt, fc, gammatone.M_BASILAR)


def perceptual_spec(data, dt, Dt, bmin=CBmin, bmax=CBmax, db=BBSZ,
			do_mod=0, do_dissonance=False, PlompBouman=True,
			do_peakalign=False):
	"""This returns something roughly like the neural signals leaving the ear.
	It filters into 1-erb-wide bins, then takes the cube root of
	the amplitude.
	@return: (channel_info, data, time_offset), where
		- C{time_offset} is the time of the the first output datum
			relative to the time of the  first input sample.
		- C{channel_info}
		- C{data}
	"""
	do_mod = int(do_mod)
	Neural_Time_Resolution = max(Neural_Tick, dt)
	m = data.shape[0]
	assert m*dt > Dt, "Tiny amount of data: %d*%g=%.2gs < Dt=%2g" % (m, dt, m*dt, Dt)
	# ntau = int(round(m*dt/Dt))
	ntau = int(M.floor(m*dt/Dt))
	# fbank = fft_filterbank(data, dt, Dt)
	fbank = gammatone_filterbank(data, dt, Dt)
	nband0 = int(M.ceil((bmax-bmin)/db))
	nbands = nband0
	assert do_mod in [0,1,2], "Do_mod=%s" % str(do_mod)
	if do_mod:
		nbands += int(do_mod)
		tk = []
	if do_dissonance:
		nbands += 1
		diss = roughness_c()
	if do_peakalign:
		nbands += 1
		palign = peakalign_c()
	neural = numpy.zeros((nbands, ntau), numpy.float)
	bctrs = []
	t0x = []
	iband = 0
	b = bmin
	while b < bmax:
		# print 'iband=', iband, 'b=', b, 'nbands=', nbands
		fc = erb_scale.erb_to_f(b)
		filtered_sig = fbank.filter(fc)
		hair_cells = numpy.maximum(filtered_sig/threshold(fc), 0.0)**(2*E)
		# pylab.plot(hair_cells)
		# pylab.title('hair cells')
		# pylab.figure()
		pwrfilt, t0 = power.smooth(hair_cells, dt, Dt)
		# pylab.plot(pwrfilt)
		t0x.append(t0)
		if PlompBouman:
			pwrfilt = numpy.maximum(VM.lopass_first_order(pwrfilt, Dt/tau_lp(fc)), 0.0)
		else:
			pwrfilt = numpy.maximum(pwrfilt, 0.0)
		# pylab.plot(pwrfilt)
		# pylab.show()
		neural[iband,:] = pwrfilt[:ntau]
		bctrs.append({'type':'band', 'smooth': tau_lp(fc), 'fc': fc,
				'erb': b, 'E': E, 'id': 'B%.1f' % b
				})

		if do_mod or do_dissonance or do_peakalign:
			tkfilt, t0 = power.smooth(hair_cells, dt, Neural_Time_Resolution)
		if do_dissonance:
			diss_bw = min(DISS_BWF*erb_scale.ebw(b), MAX_DISS_FREQ)
			diss.add(tkfilt, Neural_Time_Resolution, diss_bw)
		if do_mod:
			tk.append(tkfilt)
		if do_peakalign:
			palign.add(tkfilt, Neural_Time_Resolution, fc)
		b += db
		iband += 1
	assert iband == nband0

	if do_mod == 1:
		tksum, tkn, tkmin = process_voicing(tk, Neural_Time_Resolution, Dt)
		neural[iband,:] = (tksum/tkn-tkmin)[:ntau]*0.15
		bctrs.append({'type':'haspitch', 'id':'haspitch1', 'variant': 1,
				'Fentropy': -M.log(len(tk))
				})
		iband += 1
		assert len(bctrs) == iband
	elif do_mod == 2:
		Voicing_divide = 0.5*(CBmax+CBmin)
		tk_low = []
		tk_high = []
		for (bctr, tki) in zip(bctrs, tk):
			assert tki.shape[0] > 0
			if bctr['erb'] > Voicing_divide:
				tk_high.append(tki)
			else:
				tk_low.append(tki)
		tksum_low, tkn_low, tkmin_low = process_voicing(tk_low, Neural_Time_Resolution, Dt)
		tksum_high, tkn_high, tkmin_high = process_voicing(tk_high, Neural_Time_Resolution, Dt)
		avgmmin = (tksum_low+tksum_high)/(tkn_low+tkn_high) - numpy.minimum(tkmin_low, tkmin_high)
		neural[iband,:] = avgmmin[:ntau]*0.15
		bctrs.append({'type':'haspitch', 'id':'haspitch_all',
				'variant': 'all',
				'Fentropy': -M.log(len(tk))
				})
		iband += 1

		neural[iband,:] = (tksum_high/tkn_high-tkmin_high)[:ntau]*0.15
		bctrs.append({'type':'haspitch', 'id':'haspitch_high',
				'variant': 'high',
				'Fentropy': -M.log(len(tk_high))
				})
		iband += 1
		assert len(bctrs) == iband
	if do_dissonance:
		neural[iband,:] = diss.get(Dt)[:ntau]*20.0
		# pylab.plot(diss.get(Dt))
		# pylab.title('roughness_c')
		# pylab.show()
		bctrs.append({'type':'roughness', 'id':'roughness1',
				'variant': 1,
				'Fentropy': -M.log(diss.n)
				})
		iband += 1
	if do_peakalign:
		neural[iband,:] = palign.get(Dt)[:ntau]
		bctrs.append({'type':'peakalign', 'id':'peakalign1',
				'variant': 1,
				'Fentropy': -M.log(palign.n)
				})
		iband += 1
	assert iband == neural.shape[0], "iband=%d != shape=%s" % (iband, str(neural.shape))

	return (bctrs, neural, gpkmisc.median(t0x))



def test():
	dt = 1e-4
	d = numpy.zeros((10000,1))
	for i in range(0, 10000, 100):
		d[i,0] = 1.0
	return gpkimgclass.gpk_img({'CDELT2': dt}, d)



__version__ = "$Revision: 1.43 $"
__date__ = "$Date: 2007/03/07 23:54:40 $"


if __name__ == '__main__':
	# try:
		# import psyco
		# psyco.full()
	# except ImportError:
		# pass
	import gpkimgclass
	DT = 0.01	# Seconds.  Default output sampling interval.
	arglist = sys.argv[1:]
	Blocksize = 10.0
	arglist0 = arglist
	column = None
	signal = None
	outfile = "-"
	extrahdr = {}
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-dt':
			DT = float(arglist.pop(0))
		elif arg == '-f':
			signal = gpkimgclass.read(arglist.pop(0))
		elif arg == '-c':
			tmp = arglist.pop(0)
			try:
				column = int( tmp )
			except ValueError:
				column = tmp
		elif arg == '-b':
			Blocksize = float(arglist.pop(0))
		elif arg == '-plot':
			import pylab
		elif arg == '-test':
			signal = test()
			column = 0
		elif arg == '-o':
			outfile = arglist.pop(0)
		elif arg == '-write':
			extrahdr = dict( [q.strip().split('=', 1)
					for q in arglist.pop(0).split(';') ]
					)
		else:
			die.info("Unrecognized flag: %s" % arg)
			print __doc__
			die.exit(1)
	if arglist:
		if signal is None:
			signal = gpkimgclass.read(arglist[0])
		else:
			die.die("Usage: cannot specify name both with -f and as last argument.")

	if signal is None:
		die.info("No signal file specified.")
		print __doc__
		die.exit(1)
	try:
		signal.column(column)
	except KeyError:
		die.info("File has %d columns" % signal.n[1])
		die.die("Bad column: %s" % str(column))

	bctrs, neural, tshift = block_percep_spec(signal.column(0), signal.dt(), DT,
					do_mod=True, do_dissonance=True,
					do_peakalign=True,
					blocksize=Blocksize)
	o = numpy.transpose(neural)
	if pylab:
		pylab.matshow(numpy.transpose(o))
		pylab.show()

	hdr = signal.hdr.copy()
	hdr['program'] = sys.argv[0]
	hdr['ARGV'] = arglist0
	hdr['input_file'] = signalfile
	hdr['column'] = column
	hdr['CDELT2'] = DT
	hdr['CRPIX2'] = 1
	hdr['CRVAL2'] = signal.start() + tshift
	hdr['CRPIX1'] = 1
	hdr['CRVAL1'] = bctrs[0]['erb']
	hdr['CDELT1'] = bctrs[1]['erb']-bctrs[0]['erb']
	hdr['BITPIX'] = -32
	hdr.update( extrahdr )
	for (i,b) in enumerate(bctrs):
		hdr['TTYPE%d' % i] = b['id']
	gpkimgclass.gpk_img(hdr, o).write(outfile)



#!/usr/bin/env python


from gmisclib import erb_scale
from gmisclib import Num
from gpk_voicing import percep_spec
from gpk_voicing import fv_misc as M
SillyWidthException = M.SillyWidthException


def feature_vec(data, dt, DT,
		LF=1.0, Nsv=M.NSV, ELF=1.0,
		do_voicing=1, do_dissonance=True,
		PlompBouman=False, do_pdur=True):
	FORMANT_LOW = erb_scale.f_to_erb(200.0)
	FORMANT_HIGH = erb_scale.f_to_erb(4000.0)
	assert DT > 0.0 and float(DT)>0.0
	assert LF==1.0
	assert ELF==1.0
	assert not PlompBouman
	bmin = erb_scale.f_to_erb(80.0)
	bmax = erb_scale.f_to_erb(6000.0)
	ectrs, neural, t0 = percep_spec.perceptual_spec(data, dt, DT,
							bmin, bmax, M.DB,
							do_mod=do_voicing,
							do_dissonance=do_dissonance,
							PlompBouman=PlompBouman
							)

	band_indices = [i for (i,ec) in enumerate(ectrs) if ec['type']=='band']
	neural_b = neural.take(band_indices, axis=0)
	assert neural.shape[1]==neural.shape[1]

	neural_now = Num.average(neural_b, axis=0)	# Average over frequency.
	assert neural_now.shape[0] == neural.shape[1]
	neural_avg = Num.sum(neural_now**2)/Num.sum(neural_now)	# Average over time.
		# neural_avg is a scalar, grand average.
	Num.divide(neural, neural_avg, neural)
		# Now, we've normalized by an over-all average loudness.
	Num.divide(neural_now, neural_avg, neural_now)
		# Now, we've normalized by an over-all average loudness.

	assert neural_b.shape[0] <= neural.shape[1]
	for (i,e) in enumerate(ectrs):
		if e['type'] == 'haspitch':
			Num.divide(neural[i,:], neural_avg, neural[i,:])
		if e['type'] == 'dissonance':
			Num.divide(neural[i,:], neural_avg, neural[i,:])

	# print '# neural_avg=', neural_avg
	o = []
	descr = []
	wset = set()
	for vl in [0.06]:
		# print 'vl=', vl, type(vl), 'LF=', LF, type(LF), 'DT=', DT, type(DT)
		w = int(round(vl*LF/DT))
		if not w in wset:
			tmpo, tmpd = M.vowel(w, ectrs, neural, neural_now, Nsv,
						formant_low=FORMANT_LOW,
						formant_high=FORMANT_HIGH
						)
			o.extend(tmpo)
			descr.extend(tmpd)
			wset.add(w)
	assert len(descr)==len(o), "Descriptor mismatch"

	wset = set()
	for fl in [0.02]:
		w = int(round(fl*ELF/DT))
		if not w in wset:
			tmpo, tmpd = M.fricative(w, ectrs, neural, neural_now, Nsv)
			o.extend(tmpo)
			descr.extend(tmpd)
			wset.add(w)
	assert len(descr)==len(o), "Descriptor mismatch"

	wset = set()
	for fel in [0.04]:
		w = int(round(fel*ELF/DT))
		if not w in wset:
			tmpo, tmpd = M.fricative_edge(w, ectrs, neural, neural_now, Nsv,
						do_abs=False
						)
			o.extend(tmpo)
			descr.extend(tmpd)
			wset.add(w)
	assert len(descr)==len(o), "Descriptor mismatch"

	wset = set()
	for sel in [0.03]:
		w = int(round(sel*LF/DT))
		if not w in wset:
			tmpo, tmpd = M.spectral_entropy(w, ectrs, neural, neural_now, Nsv)
			o.extend(tmpo)
			descr.extend(tmpd)
			wset.add(w)
		assert len(descr)==len(o), "Descriptor mismatch"
	assert len(descr)==len(o), "Descriptor mismatch"
	wset = set()
	for sel in [0.07]:
		w = int(round(sel*LF/DT))
		if not w in wset:
			tmpo, tmpd = M.space_time_entropy(w, ectrs, neural, neural_now, Nsv)
			o.extend(tmpo)
			descr.extend(tmpd)
			wset.add(w)
		assert len(descr)==len(o), "Descriptor mismatch"
	assert len(descr)==len(o), "Descriptor mismatch"


	if do_voicing:
		wset = set()
		for hpl in [0.02]:
			w = int(round(hpl*LF/DT))
			if not w in wset:
				tmpo, tmpd = M.haspitch(w, ectrs, neural, neural_now, Nsv)
				o.extend(tmpo)
				descr.extend(tmpd)
				wset.add(w)
	assert len(descr)==len(o), "Descriptor mismatch"

	if do_dissonance:
		wset = set()
		for dsl in [0.06]:
			w = int(round(dsl*LF/DT))
			if not w in wset:
				tmpo, tmpd = M.dissonance(w, ectrs, neural, neural_now, Nsv)
				o.extend(tmpo)
				descr.extend(tmpd)
				wset.add(w)

	assert len(descr)==len(o), "Descriptor mismatch"
	if do_pdur:
		tmpo, tmpd = M.pdur(ectrs, neural, neural_now, DT)
		o.extend(tmpo)
		descr.extend(tmpd)

	assert len(descr)==len(o), "Descriptor mismatch"
	N = neural[0].shape[0]
	for (i, (tmp, dsc)) in enumerate(zip(o, descr)):
		assert tmp.shape == (N,), "Wrong size: %d, %s = %d vs. %d" % (i, str(dsc), tmp.shape[0], N)

	return (o, descr, DT, t0)




#!python

# This is a fancy installer that will download some missing packages.
# It's pretty good, so long as you have setuptools in your python distribution.

# _imported = False
# try:
	# from setuptools import setup
	# _imported = True
	# pass
# except ImportError:
	# pass
# 
# #If it doesn't work, you can try the line below.   This uses the
# # easy_setup.py script to download and install setuptools.
# if not _imported:
	# print "EZ_setup"
	# try:
		# from ez_setup import use_setuptools
		# use_setuptools()
		# from setuptools import setup
		# _imported = True
		# pass
	# except ImportError:
		# pass
# 
# # Finally, use distutils.   That should work, but you'll have to download
# # all the required packages by hand.
# if not _imported:
from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext

import numpy
import os
Include_dirs = [ '.', 'lib', numpy.get_include(),
			os.path.join(os.environ['HOME'], 'local', 'include') ]
Ldirs = [ os.path.join(os.environ['HOME'], 'local', 'lib') ]

setup(name="gpk-voicing", version = "1.7.2",
	description = "Python Libraries and scripts for speech analysis.",
	author = "Greg Kochanski",
	url = "http://kochanski.org/gpk/code/speechresearch/voicing",
	author_email = "gpk@kochanski.org",
	packages = ['gpk_voicing'],
	package_dir = {'gpk_voicing': 'lib'},
	scripts = ['emphasis_stevens.py', 'feature_vec.py', 'pseudoduration.py',
			'irregularity.py', 'g_audioprep.py', 'xform_feature_vec.py',
			'audio_cat_normalize.py'
			],
	ext_modules=[ 
    		Extension("gpk_voicing.power1",         ["lib/power1.pyx"],
				include_dirs = Include_dirs,
				depends = [],
				library_dirs = Ldirs,
				libraries = ['gpk'],
				language = 'c++'
				),
    		Extension("gpk_voicing.gammatone",         ["lib/gammatone.pyx"],
				include_dirs = Include_dirs,
				depends = [],
				library_dirs = Ldirs,
				libraries = ['gpk'],
				language = 'c++'
				),
    		Extension("gpk_voicing.percep_spec_extras",         ["lib/percep_spec_extras.pyx"],
				include_dirs = Include_dirs,
				depends = [],
				library_dirs = Ldirs,
				libraries = ['gpk'],
				language = 'c++'
				),
    		Extension("gpk_voicing.power1",         ["lib/power1.pyx"],
				include_dirs = Include_dirs,
				depends = [],
				library_dirs = Ldirs,
				libraries = ['gpk'],
				language = 'c++'
				),
		],
	# install_requires = [
				# "gpkavg >= 1.0",
				# "gpkimg >= 1.2",
				# "gmisclib >= 0.60.5",
				# 'numpy >= 1.0.3',
				# ],
	cmdclass = {'build_ext': build_ext},
	# extras_require = {},
	dependency_links = [
				],
	license = 'GPL2',
	keywords = ("phonetics speech computation linguistics scripts library python "
			+ "science perception feature vector loudness spectral slope"),
	platforms = "All",
	classifiers = [
		'License :: OSI Approved :: GNU General Public License (GPL)',
		'Operating System :: OS Independent',
		'Development Status :: 3 - Alpha',
		'Topic :: Scientific/Engineering',
		'Environment :: Console',
		'Intended Audience :: Science/Research',
		'Programming Language :: Python',
		# X11
		'Operating System :: OS Independent',
		'Programming Language :: Python',
		],
	long_description = """Python Libraries for speech analysis.
	See http://kochanski.plus.com/code/speechresearch/voicing for documentation
	and http://sourceforge.net/projects/speechresearch for downloads,
	or look on the Python Cheese shop, http://pypi.python.org .
	"""
	)




#!/usr/bin/env python

"""Perceptual loudness measurement.
This is an implementation of
"Percieved Level of Noise by Mark VII and Decibels (E)"
S. S. Stevens,
J. Acoustical Soc. Am. v. 51(2, part 2) 1972
pages 575-602.

The algorithm matches the paper for FractOct==ThirdOct or OneOct.
The algorithm assumes that the level defined by SIXTY_EIGHT
corresponds to 68dB relative to 20 micro-Newtons per square meter
sound pressure.

USER INFORMATION:

You'll find, somewhere, a subdirectory named
speechresearch/voicing .
Within this is a script called emphasis_stevens.py .
This can be run as
python emphasis_stevens.py -o outputfile -write BITPIX=0 -c channelnumber inputfile
(there are a couple of other flags, too).

It will then read the input file (which is an audio recording)
and produce a time-series of the loudness.   Data input and
output is via the gpkio library, in .../speechresearch/gpkio.
That library has many virtues, but reading WAV files is not
one of them, so you have to convert .wav files to
the "GPK ASCII image" format (or one of several other formats),
using  .../speechresearch/lib/wavio.py .

The output will be in an ASCII version of the format,
which should be reasonably intelligible.    (The GPK ASCII image
format is based on the FITS astronomy format from NASA.)

However, that script uses the gpkio and gpklib libraries
(also under ../speechresearch).  These need to be compiled,
and it uses the gpk_img_python package that needs to be
installed via   "python setup.py install".
Oh, and .../speechresearch/gmisclib needs to be in
your PYTHONPATH.

Give it a try, and I'll be happy to help, and will
incorporate the troubles you have into some form of
documentation.    Sorry, I have nothing better yet.
"""

import sys
import gpkimgclass
from gmisclib import Num
from gmisclib import die
from gpk_voicing import emphasis


if __name__ == '__main__':
	DT = 0.01	# Seconds.  Default output sampling interval.
	arglist = sys.argv[1:]
	columns = None
	signalfile = None
	outfile = None
	extrahdr = {}
	while len(arglist)>0 and arglist[0][0]=='-':
		arg = arglist.pop(0)
		if arg == '-c':
			if columns == None:
				columns = []
			tmp = arglist.pop(0)
			try:
				col = int(tmp)
			except ValueError:
				col = tmp
			columns.append(col)
		elif arg == '-cal':
			emphasis.SIXTY_EIGHT = float(arglist.pop(0))
		elif arg == '-f':
			signalfile = arglist.pop(0)
			die.note("signalfile", signalfile)
		elif arg == '-o':
			outfile = arglist.pop(0)
		elif arg == '-dt':
			DT = float(arglist.pop(0))
		elif arg == '-write':
			extrahdr = dict( [q.strip().split('=', 1)
					for q in arglist.pop(0).split(';') ]
					)
		else:
			die.die("Unrecognized flag: %s" % arg)
	if len(arglist) > 0:
		rootname = arglist.pop(0)
		if signalfile == None:
			signalfile = rootname + '.d'
		if outfile == None:
			outfile = rootname + '.emph.dat'
	if outfile == None:
		outfile = 'emph.dat'
	signal = gpkimgclass.read(signalfile)
	nch = signal.d.shape[1]
	if columns == None:
		columns = range(nch)
	die.note("columns", str(columns))
	for c in columns:
		try:
			signal.column(c)
		except KeyError:
			die.info("File has %d columns" % signal.n[1])
			die.die("Bad column: %s" % str(c))

	tmp = [ emphasis.simple_emphasis(signal.column(i), signal.dt(), DT)
			for i in columns ]
	n = len(tmp[0])
	o = Num.zeros((n,), Num.Float)
	for t in tmp:
		Num.add(o, t, o)
	o = o / nch

	hdr = signal.hdr.copy()
	hdr['CDELT2'] = DT
	hdr['CRPIX2'] = 0
	hdr['CRVAL2'] = signal.start()
	hdr['CDELT1'] = 1
	hdr['TTYPE1'] = 'loudness'
	hdr['BITPIX'] = -32
	hdr.update( extrahdr )
	oo = gpkimgclass.gpk_img(hdr, Num.transpose( [o]))
	oo.write(outfile)




#!/usr/bin/env python

"""This script takes the loudest channel from a multichannel
.wav file and converts it to any format known by gpkimgclass.py
"""
from gpk_voicing import dirty_io



if __name__ == '__main__':
	import sys
	channel = None
	arglist = sys.argv[1:]
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-c':
			channel = int(arglist.pop(0))
		else:
			die.die('Unrecognized argument: %s' % arg)
	if len(arglist) != 2:
		print __doc__
		sys.exit(1)

	x = dirty_io.wav_read(arglist[0], channel)
	x.write(arglist[1])
